from unittest import TestCase

import ez_yaml
import pytest
from deepdiff import DeepDiff
from unittest.mock import MagicMock

from gitlabform.constants import APPROVAL_RULE_NAME
from gitlabform import EXIT_INVALID_INPUT
from gitlabform.configuration import Configuration
from gitlabform.configuration.transform import (
    AccessLevelsTransformer,
    UserTransformer,
    ImplicitNameTransformer,
    MergeRequestApprovalsTransformer,
)
from gitlabform.gitlab import GitLab


def test__config__with_access_level_names__branches():
    config_yaml = f"""
    projects_and_groups:
      foobar/*:
        branches:
          develop:
            protected: false
          main:
            protected: true
            push_access_level: no access
            merge_access_level: developer
            unprotect_access_level: maintainer
            code_owner_approval_required: true
          branch_protected_from_changes:
            protected: true
            push_access_level: no access
            merge_access_level: no access
            unprotect_access_level: maintainer
      "*":
        branches:
          '*-something':
            protected: true
            push_access_level: no access
            merge_access_level: developer
            unprotect_access_level: maintainer
          allow_to_force_push:
            protected: true
            push_access_level: developer
            merge_access_level: developer
            unprotect_access_level: maintainer
            allow_force_push: true
    """
    configuration = Configuration(config_string=config_yaml)

    transformer = AccessLevelsTransformer(MagicMock(GitLab))
    transformer.transform(configuration)

    config_with_numbers = f"""
    projects_and_groups:
      foobar/*:
        branches:
          develop:
            protected: false
          main:
            protected: true
            push_access_level: 0
            merge_access_level: 30
            unprotect_access_level: 40
            code_owner_approval_required: true
          branch_protected_from_changes:
            protected: true
            push_access_level: 0
            merge_access_level: 0
            unprotect_access_level: 40
      "*":
        branches:
          '*-something':
            protected: true
            push_access_level: 0
            merge_access_level: 30
            unprotect_access_level: 40
          allow_to_force_push:
            protected: true
            push_access_level: 30
            merge_access_level: 30
            unprotect_access_level: 40
            allow_force_push: true
    """
    configuration_with_numbers = Configuration(config_string=config_with_numbers)

    ddiff = DeepDiff(configuration.config, configuration_with_numbers.config)
    assert not ddiff


def test__config__with_access_level_names__group_ldap_links():
    config_yaml = f"""
    projects_and_groups:
      foobar/*:
        group_ldap_links:
          # "provider" field should contain a value that you can find in the GitLab web UI,
          # see https://github.com/gitlabform/gitlabform/issues/261
          devops_are_maintainers:
            provider: "AD"
            cn: "devops"
            group_access: maintainer
          developers_are_developers:
            provider: "AD"
            filter: "(employeeType=developer)"
            group_access: developer
    """
    configuration = Configuration(config_string=config_yaml)

    transformer = AccessLevelsTransformer(MagicMock(GitLab))
    transformer.transform(configuration)

    config_with_numbers = f"""
    projects_and_groups:
      foobar/*:
        group_ldap_links:
          # "provider" field should contain a value that you can find in the GitLab web UI,
          # see https://github.com/gitlabform/gitlabform/issues/261
          devops_are_maintainers:
            provider: "AD"
            cn: "devops"
            group_access: 40
          developers_are_developers:
            provider: "AD"
            filter: "(employeeType=developer)"
            group_access: 30
    """
    configuration_with_numbers = Configuration(config_string=config_with_numbers)

    ddiff = DeepDiff(configuration.config, configuration_with_numbers.config)
    assert not ddiff


def test__config__with_access_level_names__branches_premium_syntax():
    config_yaml = f"""
    projects_and_groups:
      foobar/*:
        branches:
          special:
            protected: true
            allowed_to_push:
              - user: jsmith # you can use usernames...
              - user: bdoe
              - group: another-group # ...or group names (paths)...
            allowed_to_merge:
              - user_id: 15 # ...or user ids, if you know them...
              - access_level: developer # ...or the whole access levels, like in the other syntax
              - group_id: 456 # ...or group ids, if you know them...
            allowed_to_unprotect:
              - access_level: maintainer # ...or the whole access levels, like in the other syntax
    """
    configuration = Configuration(config_string=config_yaml)

    transformer = AccessLevelsTransformer(MagicMock(GitLab))
    transformer.transform(configuration)

    config_with_numbers = f"""
    projects_and_groups:
      foobar/*:
        branches:
          special:
            protected: true
            allowed_to_push:
              - user: jsmith # you can use usernames...
              - user: bdoe
              - group: another-group # ...or group names (paths)...
            allowed_to_merge:
              - user_id: 15 # ...or user ids, if you know them...
              - access_level: 30 # ...or the whole access levels, like in the other syntax
              - group_id: 456 # ...or group ids, if you know them...
            allowed_to_unprotect:
              - access_level: 40 # ...or the whole access levels, like in the other syntax
    """
    configuration_with_numbers = Configuration(config_string=config_with_numbers)

    ddiff = DeepDiff(configuration.config, configuration_with_numbers.config)
    assert not ddiff


def test__config__with_access_level_names__invalid_name():
    config_yaml = f"""
    projects_and_groups:
      foobar/*:
        branches:
          special:
            protected: true
            allowed_to_push:
              - user: jsmith # you can use usernames...
              - user: bdoe
              - group: another-group # ...or group names (paths)...
            allowed_to_merge:
              - user_id: 15 # ...or user ids, if you know them...
              - access_level: developers # <-------------------------- this is invalid, it's plural
              - group_id: 456 # ...or group ids, if you know them...
            allowed_to_unprotect:
              - access_level: maintainer # ...or the whole access levels, like in the other syntax
    """
    configuration = Configuration(config_string=config_yaml)
    transformer = AccessLevelsTransformer(MagicMock(GitLab))

    with pytest.raises(SystemExit) as e:
        transformer.transform(configuration)

    assert e.value.code == EXIT_INVALID_INPUT


class TestUserTransformer(TestCase):
    def test__transform_for_protected_environments(self) -> None:
        config_yaml = f"""
        projects_and_groups:
          "foo/bar":
            protected_environments:
              foo:
                name: foo
                deploy_access_levels:
                  - access_level: maintainer
                    group_inheritance_type: 0
                  - user: jsmith
                  - user_id: 123
                  - user: bdoe
                  - user_id: 456
        """
        configuration = Configuration(config_string=config_yaml)

        gitlab_mock = MagicMock(GitLab)
        gitlab_mock._get_user_id = MagicMock(side_effect=[78, 89])

        transformer = UserTransformer(gitlab_mock)
        transformer.transform(configuration)

        assert gitlab_mock._get_user_id.call_count == 2

        expected_transformed_config_yaml = f"""
        projects_and_groups:
          "foo/bar":
            protected_environments:
              foo:
                name: foo
                deploy_access_levels:
                  - access_level: maintainer
                    group_inheritance_type: 0
                  - user_id: 78
                  - user_id: 123
                  - user_id: 89
                  - user_id: 456
        """

        expected_transformed_config = Configuration(
            config_string=expected_transformed_config_yaml
        )

        assert not DeepDiff(configuration.config, expected_transformed_config.config)


class TestImplicitNameTransformer(TestCase):
    _base_cfg = f"""
        projects_and_groups:
          "foo/bar":
            protected_environments:
              enforce: true
              foo:
                deploy_access_levels:
                  - access_level: 40
        """

    def test__transform_for_protected_environments(self):
        configuration = Configuration(config_string=self._base_cfg)

        transformer = ImplicitNameTransformer(MagicMock(GitLab))
        transformer.transform(configuration)

        expected_transformed_config_yaml = f"""
        {self._base_cfg}
                name: foo
        """

        expected_transformed_config = Configuration(
            config_string=expected_transformed_config_yaml
        )

        assert not DeepDiff(configuration.config, expected_transformed_config.config)

    def test__transform_for_protected_environments_sanity_check(self):
        configuration = Configuration(config_string=self._base_cfg)

        transformer = ImplicitNameTransformer(MagicMock(GitLab))
        transformer.transform(configuration)

        expected_transformed_config_yaml = f"""
        {self._base_cfg}
                name: blah
        """

        expected_transformed_config = Configuration(
            config_string=expected_transformed_config_yaml
        )

        assert DeepDiff(configuration.config, expected_transformed_config.config)


#
# TODO: remove below test(s) in v4.x
#


class TestMergeRequestApprovalsTransformer(TestCase):
    def test__config__with_merge_request_approvals__many_users_and_groups_and_remove(
        self,
    ):
        config_before = f"""
        projects_and_groups:
          foobar/*:
            merge_requests:
              approvals:
                approvals_before_merge: 2
                reset_approvals_on_push: true
                disable_overriding_approvers_per_merge_request: true
              approvers:
                - user1
                - user2
              approver_groups:
                - my-group
                - my-group1/subgroup
                - my-group2/subgroup/subsubgroup
              remove_other_approval_rules: true
        """
        config_before = Configuration(config_string=config_before)

        transformer = MergeRequestApprovalsTransformer(MagicMock(GitLab))
        transformer.transform(config_before)
        config_before = ez_yaml.to_string(obj=config_before.config, options={})

        config_after = f"""
        projects_and_groups:
          foobar/*:
            merge_requests_approval_rules:
              legacy:
                approvals_required: 2
                name: {APPROVAL_RULE_NAME}
                users:
                  - user1
                  - user2
                groups:
                  - my-group
                  - my-group1/subgroup
                  - my-group2/subgroup/subsubgroup
              enforce: true
            merge_requests_approvals:
              reset_approvals_on_push: true
              disable_overriding_approvers_per_merge_request: true
        """
        config_after = Configuration(config_string=config_after)
        config_after = ez_yaml.to_string(obj=config_after.config, options={})

        assert config_before == config_after

    def test__config__with_merge_request_approvals__single_user_no_remove(self):
        config_before = f"""
        projects_and_groups:
          '*':
            merge_requests:
              approvals:
                approvals_before_merge: 2
                reset_approvals_on_push: false
                disable_overriding_approvers_per_merge_request: false
              approvers:
                - user1
        """
        config_before = Configuration(config_string=config_before)

        transformer = MergeRequestApprovalsTransformer(MagicMock(GitLab))
        transformer.transform(config_before)
        config_before = ez_yaml.to_string(obj=config_before.config, options={})

        config_after = f"""
        projects_and_groups:
          '*':
            merge_requests_approval_rules:
              legacy:
                approvals_required: 2
                name: {APPROVAL_RULE_NAME}
                users:
                  - user1
            merge_requests_approvals:
              reset_approvals_on_push: false
              disable_overriding_approvers_per_merge_request: false
        """
        config_after = Configuration(config_string=config_after)
        config_after = ez_yaml.to_string(obj=config_after.config, options={})

        assert config_before == config_after

    def test__config__with_merge_request_approvals__single_group_no_remove(self):
        config_before = f"""
        projects_and_groups:
          'foo/bar':
            merge_requests:
              approvals:
                approvals_before_merge: 1
                reset_approvals_on_push: false
                disable_overriding_approvers_per_merge_request: true
              approver_groups:
                - my-group2
        """
        config_before = Configuration(config_string=config_before)

        transformer = MergeRequestApprovalsTransformer(MagicMock(GitLab))
        transformer.transform(config_before)
        config_before = ez_yaml.to_string(obj=config_before.config, options={})

        config_after = f"""
        projects_and_groups:
          'foo/bar':
            merge_requests_approval_rules:
              legacy:
                approvals_required: 1
                name: {APPROVAL_RULE_NAME}
                groups:
                  - my-group2
            merge_requests_approvals:
              reset_approvals_on_push: false
              disable_overriding_approvers_per_merge_request: true
        """
        config_after = Configuration(config_string=config_after)
        config_after = ez_yaml.to_string(obj=config_after.config, options={})

        assert config_before == config_after

    def test__config__with_merge_request_approvals__guessing_approvals_before_merge(
        self,
    ):
        config_before = f"""
        projects_and_groups:
          'foo/bar':
            merge_requests:
              approvals:
                reset_approvals_on_push: false
                disable_overriding_approvers_per_merge_request: true
              approver_groups:
                - my-group2
        """
        config_before = Configuration(config_string=config_before)

        transformer = MergeRequestApprovalsTransformer(MagicMock(GitLab))
        transformer.transform(config_before)
        config_before = ez_yaml.to_string(obj=config_before.config, options={})

        config_after = f"""
        projects_and_groups:
          'foo/bar':
            merge_requests_approval_rules:
              legacy:
                approvals_required: 2
                name: {APPROVAL_RULE_NAME}
                groups:
                  - my-group2
            merge_requests_approvals:
              reset_approvals_on_push: false
              disable_overriding_approvers_per_merge_request: true
        """
        config_after = Configuration(config_string=config_after)
        config_after = ez_yaml.to_string(obj=config_after.config, options={})

        assert config_before == config_after



from gitlabform.gitlab.core import GitLabCore


class GitLabMembers(GitLabCore):
    def get_project_members(self, project_and_group_name, all=False):
        url_template = "projects/%s/members"
        if all:
            url_template += "/all"

        return self._make_requests_to_api(url_template, project_and_group_name)

    def get_shared_with_groups(self, project_and_group_name):
        # a dict with groups that this project has been shared with
        return self._make_requests_to_api("projects/%s", project_and_group_name)[
            "shared_with_groups"
        ]

    def add_member_to_project(
        self,
        project_and_group_name,
        username,
        access_level,
        expires_at=None,
        user_id=None,
    ):
        if not user_id:
            user_id = self._get_user_id(username)
        data = {"user_id": user_id, "expires_at": expires_at}
        if access_level is not None:
            data["access_level"] = access_level

        return self._make_requests_to_api(
            "projects/%s/members",
            project_and_group_name,
            method="POST",
            data=data,
            expected_codes=201,
        )

    def remove_member_from_project(
        self, project_and_group_name, username, user_id=None
    ):
        if not user_id:
            user_id = self._get_user_id(username)
        # 404 means that the user is already not a member of the project, so let's accept it for idempotency
        return self._make_requests_to_api(
            "projects/%s/members/%s",
            (project_and_group_name, user_id),
            method="DELETE",
            expected_codes=[204, 404],
        )

    def get_group_members(self, group_name, with_inherited=False):
        url_template = "groups/%s/members"
        if with_inherited:
            url_template += "/all"

        return self._make_requests_to_api(url_template, group_name)

    def get_members_from_project(self, project_and_group_name):
        # note that this DOES NOT return inherited users
        members = self._make_requests_to_api(
            "projects/%s/members", project_and_group_name
        )
        # it will return {username1: {...api info about username1...}, username2: {...}}
        # otherwise it can get very long to iterate when checking if a user
        # is already in the project if there are a lot of users to check
        final_members = {}
        for member in members:
            final_members[member["username"]] = member

        return final_members

    def add_member_to_group(
        self, group_name, username, access_level, expires_at=None, user_id=None
    ):
        if not user_id:
            user_id = self._get_user_id(username)
        data = {"user_id": user_id, "expires_at": expires_at}
        if access_level is not None:
            data["access_level"] = access_level

        return self._make_requests_to_api(
            "groups/%s/members",
            group_name,
            method="POST",
            data=data,
            expected_codes=201,
        )

    def remove_member_from_group(self, group_name, username, user_id=None):
        if not user_id:
            user_id = self._get_user_id(username)
        # 404 means that the user is already removed, so let's accept it for idempotency
        return self._make_requests_to_api(
            "groups/%s/members/%s",
            (group_name, user_id),
            method="DELETE",
            expected_codes=[204, 404],
        )



import pytest
from tests.acceptance import (
    run_gitlabform,
)


class TestRunning:
    # noinspection PyPep8Naming
    def test__ALL(self, gitlab, group, project, other_project):
        config = f"""
        projects_and_groups:
          '*':
            project_settings:
              request_access_enabled: true
        """

        run_gitlabform(config, "ALL")

        project = gitlab.get_project(f"{group}/{project}")
        assert project["request_access_enabled"] is True

        other_project = gitlab.get_project(f"{group}/{other_project}")
        assert other_project["request_access_enabled"] is True

    # noinspection PyPep8Naming
    def test__ALL_DEFINED(self, gitlab, group, project, other_project):
        group_and_project = f"{group}/{project}"

        config = f"""
        projects_and_groups:
          {group_and_project}:
            project_settings:
              suggestion_commit_message: 'foobar'
        """

        run_gitlabform(config, "ALL_DEFINED")

        project = gitlab.get_project(group_and_project)
        assert project["suggestion_commit_message"] == "foobar"

        group_and_other_project = f"{group}/{other_project}"
        project = gitlab.get_project(group_and_other_project)
        assert project["suggestion_commit_message"] != "foobar"

        config = f"""
        projects_and_groups:
          non/existent_project:
            project_settings:
              suggestion_commit_message: 'foobar'
        """

        with pytest.raises(SystemExit):
            run_gitlabform(config, "ALL_DEFINED")

        config = f"""
        projects_and_groups:
          non_existent_group/*:
            project_settings:
              suggestion_commit_message: 'foobar'
        """

        with pytest.raises(SystemExit):
            run_gitlabform(config, "ALL_DEFINED")



import logging
import pytest

from gitlabform.configuration import Configuration

logger = logging.getLogger(__name__)


@pytest.fixture
def configuration_with_subgroups_and_projects():
    config_yaml = """
    ---
    projects_and_groups:
      some_group/*:
        project_settings:
          setting_from_group: foo

      some_group/subgroup_level_1/*:
        project_settings:
          setting_from_subgroup_level_1: foo

      some_group/subgroup_level_1/some_project:
        project_settings:
          setting_from_project: foo

      some_group/subgroup_level_1/subgroup_level_2/*:
        project_settings:
          setting_from_subgroup_level_2: foo

      some_group/subgroup_level_1/subgroup_level_2/some_project:
          project_settings:
            setting_from_project: foo
    """
    return Configuration(config_string=config_yaml)


def test__get_effective_config_for_project__level1(
    configuration_with_subgroups_and_projects,
):
    effective_config = (
        configuration_with_subgroups_and_projects.get_effective_config_for_project(
            "some_group/subgroup_level_1/some_project"
        )
    )

    additive__project_settings = effective_config["project_settings"]

    # for project settings all the levels up are inherited
    assert additive__project_settings == {
        "setting_from_group": "foo",
        "setting_from_subgroup_level_1": "foo",
        "setting_from_project": "foo",
    }


def test__get_effective_config_for_project__level2(
    configuration_with_subgroups_and_projects,
):
    effective_config = (
        configuration_with_subgroups_and_projects.get_effective_config_for_project(
            "some_group/subgroup_level_1/subgroup_level_2/some_project"
        )
    )

    additive__project_settings = effective_config["project_settings"]

    # the settings should be inherited from all levels
    assert additive__project_settings == {
        "setting_from_group": "foo",
        "setting_from_subgroup_level_1": "foo",
        "setting_from_subgroup_level_2": "foo",
        "setting_from_project": "foo",
    }


@pytest.fixture
def configuration_with_subgroups_only():
    config_yaml = """
    ---
    projects_and_groups:
      some_group/*:
        group_settings:
          project_creation_level: maintainer
          subgroup_creation_level: owner
          visibility: internal

      some_group/subgroup/*:
        group_settings:
          project_creation_level: developer
            
      some_group/subgroup/subsubgroup/*:
        group_settings:
          visibility: private
    """
    return Configuration(config_string=config_yaml)


def test__get_effective_config_for_group__level1(configuration_with_subgroups_only):
    effective_config = configuration_with_subgroups_only.get_effective_config_for_group(
        "some_group/subgroup"
    )

    # the settings should be inherited from all levels
    assert effective_config == {
        "group_settings": {
            "project_creation_level": "developer",
            "subgroup_creation_level": "owner",
            "visibility": "internal",
        },
    }


def test__get_effective_config_for_group__level2(configuration_with_subgroups_only):
    effective_config = configuration_with_subgroups_only.get_effective_config_for_group(
        "some_group/subgroup/subsubgroup"
    )

    # the settings should be inherited from all levels
    assert effective_config == {
        "group_settings": {
            "project_creation_level": "developer",
            "subgroup_creation_level": "owner",
            "visibility": "private",
        },
    }



from typing import Tuple

from cli_ui import fatal

from gitlabform.constants import EXIT_INVALID_INPUT
from gitlabform.lists import OmissionReason, Groups, Projects
from gitlabform.lists.groups import GroupsProvider

from gitlabform.gitlab.core import NotFoundException


class ProjectsProvider(GroupsProvider):
    """
    For a query like "project/group", "group", "group/subgroup", ALL or ALL_DEFINED this
    class gets the effective lists of projects, taking into account skipped projects
    and the fact that the group and project names case are somewhat case-sensitive.

    Because the projects depend on groups requested, this class inherits GroupsProvider.
    """

    def __init__(self, gitlab, configuration, include_archived_projects):
        super().__init__(gitlab, configuration)
        self.include_archived_projects = include_archived_projects

    def get_projects(self, target: str) -> Projects:
        """
        :param target: "project/group", "group", "group/subgroup", ALL or ALL_DEFINED
        :return: Projects
        """

        groups = self.get_groups(target)

        if target not in ["ALL", "ALL_DEFINED"]:
            if len(groups.get_effective()) == 1:
                projects = self._get_projects(target, groups)
            else:
                projects = self._get_single_project(target)
        else:
            projects = self._get_projects(target, groups)

        return projects

    def _get_single_project(self, target: str) -> Projects:

        projects = Projects()

        # it may be a single project
        try:
            maybe_project = self.gitlab.get_project_case_insensitive(target)
            projects.add_requested([maybe_project["path_with_namespace"]])

        except NotFoundException:
            # it's a group or a subgroup - ignore it here
            pass

        return projects

    def _get_projects(self, target: str, groups: Groups) -> Projects:

        projects = Projects()

        # the source of projects are the *effective* requested groups
        (
            projects_from_groups,
            archived_projects_from_groups,
        ) = self._get_all_and_archived_projects_from_groups(groups.get_effective())
        projects.add_requested(projects_from_groups)
        projects.add_omitted(OmissionReason.ARCHIVED, archived_projects_from_groups)

        if target == "ALL_DEFINED":

            # in this case we also need to get the list of projects explicitly
            # defined in the configuration

            projects_from_configuration = self.configuration.get_projects()

            # ...but we don't need to re-check for being archived projects that we
            # already got from groups

            # TODO: this check should be case-insensitive
            projects_from_configuration_not_from_groups = [
                project
                for project in projects_from_configuration
                if project not in projects.requested
            ]

            archived_projects_from_configuration_not_from_groups = (
                self._verify_if_projects_exist_and_get_archived_projects(
                    projects_from_configuration_not_from_groups
                )
            )

            projects.add_requested(projects_from_configuration_not_from_groups)
            projects.add_omitted(
                OmissionReason.ARCHIVED,
                archived_projects_from_configuration_not_from_groups,
            )

        # TODO: consider checking for skipped earlier to avoid making requests for projects that will be skipped anyway
        projects.add_omitted(
            OmissionReason.SKIPPED, self._get_skipped_projects(projects.get_effective())
        )

        return projects

    def _verify_if_projects_exist_and_get_archived_projects(
        self, projects: list
    ) -> list:
        archived = []
        for project in projects:
            try:
                project_object = self.gitlab.get_project_case_insensitive(project)
                if project_object["archived"]:
                    archived.append(project_object["path_with_namespace"])
            except NotFoundException:
                fatal(
                    f"Configuration contains project {project} but it cannot be found in GitLab!",
                    exit_code=EXIT_INVALID_INPUT,
                )
        return archived

    def _get_all_and_archived_projects_from_groups(
        self, groups: list
    ) -> Tuple[list, list]:
        all = []
        archived = []
        for group in groups:
            if self.include_archived_projects:
                all += self.gitlab.get_projects(group, include_archived=True)
            else:
                project_objects = self.gitlab.get_projects(
                    group, include_archived=True, only_names=False
                )
                for project_object in project_objects:
                    project = project_object["path_with_namespace"]
                    all.append(project)
                    if project_object["archived"]:
                        archived.append(project)

        # deduplicate as we may have a group X and its subgroup X/Y in the groups list so the effective projects
        # may occur more than once
        return list(set(all)), list(set(archived))

    def _get_skipped_projects(self, projects: list) -> list:
        skipped = []
        for project in projects:
            if self.configuration.is_project_skipped(project):
                skipped.append(project)

        return skipped



import logging
import pytest

from gitlabform.configuration import Configuration
from gitlabform import EXIT_INVALID_INPUT

logger = logging.getLogger(__name__)


class TestInheritanceBreakValidation:
    def test__validate_break_inheritance_flag__invalid_flag_set_at_common_level(self):
        config_yaml = """
        ---
        projects_and_groups:
          "*":
            inherit: false
            secret_variables:
              secret:
                key: foo
                value: bar
        """

        configuration = Configuration(config_string=config_yaml)
        with pytest.raises(SystemExit) as exception:
            configuration.get_common_config()
        assert exception.type == SystemExit
        assert exception.value.code == EXIT_INVALID_INPUT

    def test__validate_break_inheritance_flag__invalid_flag_set_at_project_level(self):
        config_yaml = """
        ---
        projects_and_groups:
          "some_group/some_project":
            secret_variables:
              inherit: false
              secret:
                key: foo
                value: bar
        """

        configuration = Configuration(config_string=config_yaml)
        with pytest.raises(SystemExit) as exception:
            configuration.get_effective_config_for_project("some_group/some_project")
        assert exception.type == SystemExit
        assert exception.value.code == EXIT_INVALID_INPUT

    def test__validate_break_inheritance_flag__get_config_for_group__invalid_flag_set_at_group_level(
        self,
    ):
        config_yaml = """
        ---
        projects_and_groups:
          some_group/*:
            group_members:
              enforce: true
              users:
                inherit: false
                my-user:
                  access_level: maintainer
        """

        configuration = Configuration(config_string=config_yaml)
        with pytest.raises(SystemExit) as exception:
            configuration.get_effective_config_for_group("some_group")
        assert exception.type == SystemExit
        assert exception.value.code == EXIT_INVALID_INPUT

    def test__validate_break_inheritance_flag__get_config_for_project__invalid_flag_set_at_group_level(
        self,
    ):
        config_yaml = """
        ---
        projects_and_groups:
          some_group/*:
            members:
              enforce: true
              users:
                inherit: false
                my-user:
                  access_level: maintainer
        """

        configuration = Configuration(config_string=config_yaml)
        with pytest.raises(SystemExit) as exception:
            configuration.get_effective_config_for_project("some_group/*")
        assert exception.type == SystemExit
        assert exception.value.code == EXIT_INVALID_INPUT

    def test__validate_break_inheritance_flag__get_config_for_group__invalid_flag_set_at_subgroup_level(
        self,
    ):
        config_yaml = """
        ---
        projects_and_groups:
          some_group/subgroup1/*:
            members:
              enforce: true
              users:
                inherit: false
                my-user:
                  access_level: maintainer
        """

        configuration = Configuration(config_string=config_yaml)
        with pytest.raises(SystemExit) as exception:
            configuration.get_effective_config_for_group("some_group/subgroup1")
        assert exception.type == SystemExit
        assert exception.value.code == EXIT_INVALID_INPUT

    def test__validate_break_inheritance_flag__get_config_for_project__invalid_flag_set_at_subgroup_level(
        self,
    ):
        config_yaml = """
        ---
        projects_and_groups:
          some_group/subgroup1/subgroup2/*:
            members:
              enforce: true
              users:
                inherit: false
                my-user:
                  access_level: maintainer
        """

        configuration = Configuration(config_string=config_yaml)
        with pytest.raises(SystemExit) as exception:
            configuration.get_effective_config_for_project(
                "some_group/subgroup1/subgroup2/*"
            )
        assert exception.type == SystemExit
        assert exception.value.code == EXIT_INVALID_INPUT



import enum

from typing import List

from gitlabform.gitlab.branches import GitLabBranches
from gitlabform.gitlab.commits import GitLabCommits
from gitlabform.gitlab.group_badges import GitLabGroupBadges
from gitlabform.gitlab.group_ldap_links import GitLabGroupLDAPLinks
from gitlabform.gitlab.group_variables import GitLabGroupVariables
from gitlabform.gitlab.members import GitLabMembers
from gitlabform.gitlab.merge_requests import GitLabMergeRequests
from gitlabform.gitlab.pipelines import GitLabPipelines
from gitlabform.gitlab.project_badges import GitLabProjectBadges
from gitlabform.gitlab.project_deploy_keys import GitLabProjectDeployKeys
from gitlabform.gitlab.project_protected_environments import (
    GitLabProjectProtectedEnvironments,
)
from gitlabform.gitlab.project_merge_requests_approvals import (
    GitLabProjectMergeRequestsApprovals,
)
from gitlabform.gitlab.variables import GitLabVariables
from gitlabform.gitlab.repositories import GitLabRepositories
from gitlabform.gitlab.resource_groups import GitLabResourceGroups
from gitlabform.gitlab.schedules import GitLabPipelineSchedules
from gitlabform.gitlab.integrations import GitLabIntegrations
from gitlabform.gitlab.tags import GitLabTags
from gitlabform.gitlab.users import GitLabUsers


@enum.unique
class AccessLevel(enum.IntEnum):
    NO_ACCESS = 0
    MINIMAL = 5  # introduced in GitLab 13.5
    GUEST = 10
    REPORTER = 20
    DEVELOPER = 30
    MAINTAINER = 40
    OWNER = 50  # only for groups
    ADMIN = 60

    @classmethod
    def group_levels(cls) -> List[int]:
        return [level.value for level in AccessLevel if level <= 50]

    @classmethod
    def get_value(cls, name: str) -> int:
        # for the above set of key names this is enough for an effectively fuzzy name matching
        return AccessLevel[name.strip().upper().replace(" ", "_")].value

    @classmethod
    def get_canonical_names(cls) -> List[str]:
        return [level.name.lower().replace("_", " ") for level in AccessLevel]


class GitLab(
    GitLabBranches,
    GitLabCommits,
    GitLabMergeRequests,
    GitLabRepositories,
    GitLabResourceGroups,
    GitLabIntegrations,
    GitLabTags,
    GitLabGroupLDAPLinks,
    GitLabGroupBadges,
    GitLabGroupVariables,
    GitLabPipelines,
    GitLabMembers,
    GitLabUsers,
    GitLabPipelineSchedules,
    GitLabProjectBadges,
    GitLabProjectDeployKeys,
    GitLabProjectProtectedEnvironments,
    GitLabProjectMergeRequestsApprovals,
    GitLabVariables,
):
    pass



from abc import ABC

import functools
from logging import debug

from gitlabform.configuration import ConfigurationCommon
from gitlabform.util import to_str


class ConfigurationGroups(ConfigurationCommon, ABC):
    """
    Gets the groups, skipped groups and their effective configuration.

    ConfigurationCommon is an ancestor of this class because the effective group configuration
    depends on the common configuration.
    """

    def get_groups(self) -> list:
        """
        :return: sorted list of groups that are EXPLICITLY defined in the config
        """
        groups = []
        projects_and_groups = self.get("projects_and_groups")
        for element in projects_and_groups.keys():
            if element.endswith("/*"):
                # cut off that "/*"
                group_name = element[:-2]
                groups.append(group_name)
        return sorted(groups)

    def is_group_skipped(self, group):
        """
        :return: if group is defined in the key with groups to skip,
                 ignoring the case
        """
        return self._is_skipped_case_insensitively(self.get("skip_groups", []), group)

    @functools.lru_cache()
    def get_effective_config_for_group(self, group) -> dict:
        """
        :param group: "group_name"
        :return: merged configuration for this group, from common, group. Merging is additive.
        """

        common_config = self.get_common_config()
        debug("*Effective* common config: %s", to_str(common_config))

        if "/" in group:
            group_config = self._get_effective_subgroup_config(group)
        else:
            group_config = self._get_group_config(group)
        debug("*Effective* group/subgroup config: %s", to_str(group_config))

        if not common_config and group_config:
            self._validate_break_inheritance_flag(group_config, group)

        effective_config_for_group = self._merge_configs(common_config, group_config)
        debug(
            "*Effective* config common+group/subgroup: %s",
            to_str(effective_config_for_group),
        )

        return effective_config_for_group

    def _get_effective_subgroup_config(self, subgroup):

        #
        # Goes through a subgroups hierarchy, from top to bottom
        #
        # "x/y/x" -> ["x", "x/y", "x/y/z"]
        #
        # ...and for each element after 1st generate effective config from previous effective one merged with current:
        #
        #              |     v       |
        #              \---> a       |
        #                    |       v
        #                    \------>b = effective config to return
        #
        # ...where a = merged_config("x", "x/y") and b = merged_config(a, "x/y/z")
        #

        effective_config = {}
        elements = subgroup.split("/")
        last_element = None
        for element in elements:
            if not last_element:
                effective_config = self._get_group_config(element)
                debug(
                    "First level config for '%s': %s", element, to_str(effective_config)
                )
                last_element = element
            else:
                next_level_subgroup = last_element + "/" + element
                next_level_subgroup_config = self._get_group_config(next_level_subgroup)
                debug(
                    "Config for '%s': %s",
                    next_level_subgroup,
                    to_str(next_level_subgroup_config),
                )

                if effective_config:
                    self._validate_break_inheritance_flag(effective_config, subgroup)
                elif not effective_config and next_level_subgroup_config:
                    self._validate_break_inheritance_flag(
                        next_level_subgroup_config, next_level_subgroup
                    )

                effective_config = self._merge_configs(
                    effective_config,
                    next_level_subgroup_config,
                )
                debug(
                    "Merged previous level config for '%s' with config for '%s': %s",
                    last_element,
                    next_level_subgroup,
                    to_str(effective_config),
                )
                last_element = last_element + "/" + element

        return effective_config

    def _get_group_config(self, group) -> dict:
        """
        :param group: group/subgroup
        :return: configuration for this group/subgroup or empty dict if not defined,
                 ignoring the case
        """
        return self._get_case_insensitively(
            self.get("projects_and_groups"), f"{group}/*"
        )



import pytest

from gitlabform.gitlab import AccessLevel
from tests.acceptance import (
    run_gitlabform,
)


@pytest.fixture(scope="function")
def one_owner(gitlab, group, groups, sub_group, users):

    gitlab.add_member_to_group(group, users[0], AccessLevel.OWNER.value)
    gitlab.remove_member_from_group(group, "root")

    yield group

    # we are running tests with root's token, so every group is created
    # with a single user - root as owner. we restore the group to
    # this state here.

    gitlab.add_member_to_group(group, "root", AccessLevel.OWNER.value)

    # we try to remove all users, not just those added above,
    # on purpose, as more may have been added in the tests
    for user in users:
        gitlab.remove_member_from_group(group, user)

    gitlab.remove_share_from_group(group, sub_group)
    for share_with in groups:
        gitlab.remove_share_from_group(group, share_with)


class TestGroupMembersGroups:
    def test__add_group(self, gitlab, group, users, groups, one_owner):
        no_of_members_before = len(gitlab.get_group_members(group))

        add_shared_with = f"""
        projects_and_groups:
          {group}/*:
            group_members:
              users:
                {users[0]}:
                  access_level: {AccessLevel.OWNER.value}
              groups:
                {groups[0]}:
                  group_access: {AccessLevel.DEVELOPER.value}
                {groups[1]}:
                  group_access: {AccessLevel.DEVELOPER.value}
        """

        run_gitlabform(add_shared_with, group)

        members = gitlab.get_group_members(group)
        assert len(members) == no_of_members_before, members

        shared_with = gitlab.get_group_shared_with(group)
        assert len(shared_with) == 2

    def test__sub_group(self, gitlab, group, users, sub_group, one_owner):
        no_of_members_before = len(gitlab.get_group_members(group))

        add_shared_with = f"""
        projects_and_groups:
          {group}/*:
            group_members:
              users:
                {users[0]}:
                  access_level: {AccessLevel.OWNER.value}
              groups:
                {sub_group}:
                  group_access: {AccessLevel.DEVELOPER.value}
        """

        run_gitlabform(add_shared_with, group)

        members = gitlab.get_group_members(group)
        assert len(members) == no_of_members_before, members

        shared_with = gitlab.get_group_shared_with(group)
        assert len(shared_with) == 1

        # test second run (issue #236)
        run_gitlabform(add_shared_with, group)

    def test__remove_group(self, gitlab, group, users, groups, one_owner):

        gitlab.add_share_to_group(group, groups[0], AccessLevel.OWNER.value)
        gitlab.add_share_to_group(group, groups[1], AccessLevel.OWNER.value)

        no_of_members_before = len(gitlab.get_group_members(group))
        no_of_shared_with_before = len(gitlab.get_group_shared_with(group))

        remove_group = f"""
        projects_and_groups:
          {group}/*:
            group_members:
              enforce: true
              users:
                {users[0]}:
                  access_level: {AccessLevel.OWNER.value}
              groups:
                {groups[0]}:
                  group_access: {AccessLevel.DEVELOPER.value}
        """

        run_gitlabform(remove_group, group)

        members = gitlab.get_group_members(group)
        assert len(members) == no_of_members_before

        shared_with = gitlab.get_group_shared_with(group)
        assert len(shared_with) == no_of_shared_with_before - 1

        assert [sw["group_name"] for sw in shared_with] == [groups[0]]

    def test__not_remove_groups_with_enforce_false(
        self, gitlab, group, users, groups, one_owner
    ):

        no_of_members_before = len(gitlab.get_group_members(group))
        no_of_shared_with_before = len(gitlab.get_group_shared_with(group))

        setups = [
            # flag explicitly set to false
            f"""
            projects_and_groups:
              {group}/*:
                group_members:
                  enforce: false
                  users:
                    {users[0]}:
                      access_level: {AccessLevel.OWNER.value}
            """,
            # flag not set at all (but the default is false)
            f"""
            projects_and_groups:
              {group}/*:
                group_members:
                  users:
                    {users[0]}:
                      access_level: {AccessLevel.OWNER.value}
            """,
        ]
        for setup in setups:
            run_gitlabform(setup, group)

            members = gitlab.get_group_members(group)
            assert len(members) == no_of_members_before

            members_usernames = {member["username"] for member in members}
            assert members_usernames == {
                f"{users[0]}",
            }

            shared_with = gitlab.get_group_shared_with(group)
            assert len(shared_with) == no_of_shared_with_before

    def test__change_group_access(self, gitlab, group, groups, users, one_owner):

        change_some_users_access = f"""
        projects_and_groups:
          {group}/*:
            group_members:
              users:
                {users[0]}:
                  access_level: {AccessLevel.OWNER.value}
              groups:
                {groups[0]}:
                  group_access: {AccessLevel.DEVELOPER.value}
                {groups[1]}:
                  group_access: {AccessLevel.OWNER.value}
        """

        run_gitlabform(change_some_users_access, group)

        shared_with = gitlab.get_group_shared_with(group)
        for shared_with_group in shared_with:
            if shared_with_group["group_name"] == f"{groups[0]}":
                assert (
                    shared_with_group["group_access_level"]
                    == AccessLevel.DEVELOPER.value
                )
            if shared_with_group["group_name"] == f"{groups[1]}":
                assert (
                    shared_with_group["group_access_level"] == AccessLevel.OWNER.value
                )

    def test__remove_all(self, gitlab, group, users, one_owner):
        no_shared_with = f"""
        projects_and_groups:
          {group}/*:
            group_members:
              enforce: true
              users:
                {users[0]}:
                  access_level: {AccessLevel.OWNER.value}
        """

        run_gitlabform(no_shared_with, group)

        shared_with = gitlab.get_group_shared_with(group)
        assert len(shared_with) == 0

    def test__add_group_with_access_level_names(
        self, gitlab, group, users, groups, one_owner
    ):
        no_of_members_before = len(gitlab.get_group_members(group))

        add_shared_with = f"""
        projects_and_groups:
          {group}/*:
            group_members:
              users:
                {users[0]}:
                  access_level: owner
              groups:
                {groups[0]}:
                  group_access: developer
                {groups[1]}:
                  group_access: developer
        """

        run_gitlabform(add_shared_with, group)

        members = gitlab.get_group_members(group)
        assert len(members) == no_of_members_before, members

        shared_with = gitlab.get_group_shared_with(group)
        assert len(shared_with) == 2



from tests.acceptance import (
    run_gitlabform,
)


class TestDeployKeysAllProjects:
    def test__deploy_key_to_all_projects(
        self, gitlab, group, project, other_project, public_ssh_key
    ):
        deploy_key_to_all_projects = f"""
        projects_and_groups:
          "*":
            deploy_keys:
              foobar:
                key: {public_ssh_key}
                title: common_key
                can_push: false
        """
        run_gitlabform(deploy_key_to_all_projects, group)

        deploy_keys1 = gitlab.get_deploy_keys(f"{group}/{project}")
        assert len(deploy_keys1) == 1

        deploy_keys2 = gitlab.get_deploy_keys(f"{group}/{other_project}")
        assert len(deploy_keys2) == 1

    def test__deploy_key_with_spaces_in_comment_to_all_projects(
        self, gitlab, group, project, other_project, public_ssh_key
    ):
        public_ssh_key = f"{public_ssh_key} this is a comment with spaces"

        deploy_key_to_all_projects = f"""
        projects_and_groups:
          "*":
            deploy_keys:
              foobar:
                key: {public_ssh_key}
                title: common_key
                can_push: false
        """
        run_gitlabform(deploy_key_to_all_projects, group)

        deploy_keys1 = gitlab.get_deploy_keys(f"{group}/{project}")
        assert len(deploy_keys1) == 1

        deploy_keys2 = gitlab.get_deploy_keys(f"{group}/{other_project}")
        assert len(deploy_keys2) == 1



import pytest

from gitlabform import EXIT_INVALID_INPUT
from gitlabform.configuration import Configuration


def test__config_with_different_case_group():
    group_name_with_varying_case = "GROUPnameWITHvaryingCASE"
    config_yaml = f"""
    projects_and_groups:
      {group_name_with_varying_case}/*: 
        project_settings:
          visibility: internal
    """
    configuration = Configuration(config_string=config_yaml)

    group_name_with_other_case = group_name_with_varying_case.lower()

    effective_configuration = configuration.get_effective_config_for_group(
        group_name_with_other_case
    )

    assert effective_configuration["project_settings"] == {"visibility": "internal"}


def test__config_with_different_case_project():
    group_and_project_name_with_varying_case = (
        "GroupNameWithVaryingCase/projectwithvaryingcase"
    )
    config_yaml = f"""
    projects_and_groups:
      {group_and_project_name_with_varying_case}: 
        project_settings:
          visibility: public
    """
    configuration = Configuration(config_string=config_yaml)

    group_and_project_name_with_other_case = (
        group_and_project_name_with_varying_case.upper()
    )

    effective_configuration = configuration.get_effective_config_for_project(
        group_and_project_name_with_other_case
    )

    assert effective_configuration["project_settings"] == {"visibility": "public"}


def test__config_with_different_case_duplicate_groups():
    config_yaml = """
    projects_and_groups:
      groupnamewithvaryingcase/*:
        project_settings:
          visibility: internal
      GROUPnameWITHvaryingCASE/*: # different case than defined above 
        project_settings:
          visibility: public
    """

    with pytest.raises(SystemExit) as e:
        Configuration(config_string=config_yaml)
    assert e.value.code == EXIT_INVALID_INPUT


def test__config_with_different_case_duplicate_projects():
    config_yaml = """
    projects_and_groups:
      GroupNameWithVaryingCase/projectwithvaryingcase:
        project_settings:
          visibility: internal
      GroupNameWithVaryingCase/ProjectWithVaryingCase:
        project_settings:
          visibility: public
    """

    with pytest.raises(SystemExit) as e:
        Configuration(config_string=config_yaml)
    assert e.value.code == EXIT_INVALID_INPUT


def test__config_with_different_case_duplicate_skip_groups():
    config_yaml = """
    skip_groups:
      - groupnamewithvaryingcase
      - GROUPnameWITHvaryingCASE
    """

    with pytest.raises(SystemExit) as e:
        Configuration(config_string=config_yaml)
    assert e.value.code == EXIT_INVALID_INPUT


def test__config_with_different_case_duplicate_skip_projects():
    config_yaml = """
    skip_projects:
      - GroupNameWithVaryingCase/projectwithvaryingcase
      - GroupNameWithVaryingCase/ProjectWithVaryingCase
    """

    with pytest.raises(SystemExit) as e:
        Configuration(config_string=config_yaml)
    assert e.value.code == EXIT_INVALID_INPUT



from tests.acceptance import (
    run_gitlabform,
)


class TestArchiveProject:
    def test__archive_project(self, gitlab, group_and_project):

        config = f"""
        projects_and_groups:
          {group_and_project}:
            project:
              archive: true
        """

        run_gitlabform(config, group_and_project)
        project = gitlab.get_project(group_and_project)
        assert project["archived"] is True

    def test__unarchive_project(
        self, gitlab, group_and_project, other_group, other_project
    ):

        archive_project = f"""
        projects_and_groups:
          {group_and_project}:
            project:
              archive: true
        """

        run_gitlabform(archive_project, group_and_project)
        project = gitlab.get_project(group_and_project)
        assert project["archived"] is True

        unarchive_project = f"""
        projects_and_groups:
          {group_and_project}:
            project:
              archive: false
        """

        # 1. if you run gitlabform ALL, but without '--include-archived-projects',
        # then nothing will happen here as the archived project will be omitted when in
        # the effective list of groups and projects

        run_gitlabform(unarchive_project, "ALL", False)
        project = gitlab.get_project(group_and_project)
        assert project["archived"] is True

        # 2. only after you run gitlabform ALL, with '--include-archived-projects'
        # (OR pointing to it that project)
        # the target project will be unarchived
        run_gitlabform(unarchive_project, "ALL")
        project = gitlab.get_project(group_and_project)
        assert project["archived"] is False

    def test__dont_edit_archived_project(self, gitlab, group, project):
        group_and_project = f"{group}/{project}"

        archive_project = f"""
        projects_and_groups:
          {group_and_project}:
            project:
              archive: true
        """

        run_gitlabform(archive_project, group_and_project)
        project = gitlab.get_project(group_and_project)
        assert project["archived"] is True

        edit_archived_project = f"""
        # the project has to be configured as archived
        # for other configs for it to be ignored
        projects_and_groups:
          {group_and_project}:
            project:
              archive: true

          {group}/*:
            files:
              README.md:
                overwrite: true
                branches:
                  - main
                content: |
                  Some other content that the default one
        """

        run_gitlabform(edit_archived_project, group_and_project)

        # the fact that we are not getting an exception because of trying to edit
        # an archived project means that the test is passing



from gitlabform.gitlab.core import GitLabCore


def to_string(x):
    if type(x) is bool:
        # python bool is uppercase and we need a lowercase here
        return "true" if x else "false"
    else:
        return x


class GitLabGroupVariables(GitLabCore):
    def get_group_variables(self, group):
        return self._make_requests_to_api("groups/%s/variables", group)

    def post_group_variable(self, group, variable):
        # variable has to be like documented at:
        # https://docs.gitlab.com/ee/api/group_level_variables.html#create-variable

        # workaround for the GitLab bug
        variable = {k: to_string(v) for k, v in variable.items()}

        self._make_requests_to_api(
            "groups/%s/variables", group, "POST", variable, expected_codes=201
        )

    def delete_group_variable(self, group, variable_object):
        self._make_requests_to_api(
            "groups/%s/variables/%s",
            (group, variable_object["key"]),
            "DELETE",
            expected_codes=[204, 404],
        )

    def put_group_variable(self, group, variable_in_gitlab, variable_in_config):
        # variable has to be like documented at:
        # https://docs.gitlab.com/ee/api/group_level_variables.html#create-variable

        # workaround for the GitLab bug
        variable_in_config = {k: to_string(v) for k, v in variable_in_config.items()}

        self._make_requests_to_api(
            "groups/%s/variables/%s",
            (group, variable_in_gitlab["key"]),
            "PUT",
            variable_in_config,
        )

    def get_group_variable(self, group, variable_key):
        return self._make_requests_to_api(
            "groups/%s/variables/%s", (group, variable_key)
        )["value"]

    def get_group_variable_object(self, group, variable_key):
        return self._make_requests_to_api(
            "groups/%s/variables/%s", (group, variable_key)
        )



from tests.acceptance import (
    run_gitlabform,
)


class TestBadges:
    def test__badges_add(self, gitlab, group_and_project):

        config = f"""
        projects_and_groups:
          {group_and_project}:
            badges:
              pipeline-status:
                name: "Project Badge"
                link_url: "https://gitlab.example.com/%{{project_path}}/-/commits/%{{default_branch}}/foo"
                image_url: "https://gitlab.example.com/%{{project_path}}/badges/%{{default_branch}}/pipeline.svg"
        """
        run_gitlabform(config, group_and_project)

        badges = gitlab.get_project_badges(group_and_project)
        assert len(badges) == 1
        assert badges[0]["name"] == "Project Badge"

    def test__badges_delete(self, gitlab, group_and_project):

        config = f"""
            projects_and_groups:
              {group_and_project}:
                badges:
                  pipeline-status:
                    name: "Project Badge"
                    link_url: "https://gitlab.example.com/%{{project_path}}/-/commits/%{{default_branch}}/foo"
                    image_url: "https://gitlab.example.com/%{{project_path}}/badges/%{{default_branch}}/pipeline.svg"
            """
        run_gitlabform(config, group_and_project)

        badges = gitlab.get_project_badges(group_and_project)
        assert len(badges) == 1
        assert badges[0]["name"] == "Project Badge"

        config = f"""
            projects_and_groups:
              {group_and_project}:
                badges:
                  pipeline-status:
                    name: "Project Badge"
                    delete: true
            """
        run_gitlabform(config, group_and_project)

        badges = gitlab.get_project_badges(group_and_project)
        assert len(badges) == 0

    def test__badges_update(self, gitlab, group_and_project):

        config = f"""
            projects_and_groups:
              {group_and_project}:
                badges:
                  pipeline-status:
                    name: "Project Badge"
                    link_url: "https://gitlab.example.com/foo"
                    image_url: "https://gitlab.example.com/pipeline.svg"
            """
        run_gitlabform(config, group_and_project)

        badges = gitlab.get_project_badges(group_and_project)
        assert len(badges) == 1
        assert badges[0]["link_url"].endswith("foo")

        config = f"""
            projects_and_groups:
              {group_and_project}:
                badges:
                  pipeline-status:
                    name: "Project Badge"
                    link_url: "https://gitlab.example.com/bar"
                    image_url: "https://gitlab.example.com/pipeline.svg"
            """
        run_gitlabform(config, group_and_project)

        badges = gitlab.get_project_badges(group_and_project)
        assert len(badges) == 1
        assert badges[0]["link_url"].endswith("bar")

    def test__badges_update_choose_the_right_one(self, gitlab, group_and_project):

        config = f"""
            projects_and_groups:
              {group_and_project}:
                badges:
                  pipeline-status:
                    name: "Project Badge"
                    link_url: "https://gitlab.example.com/first"
                    image_url: "https://gitlab.example.com/first"
                  another:
                    name: "Project Badge 2"
                    link_url: "https://gitlab.example.com/second"
                    image_url: "https://gitlab.example.com/second" 
            """
        run_gitlabform(config, group_and_project)

        badges = gitlab.get_project_badges(group_and_project)
        assert len(badges) == 2

        config = f"""
            projects_and_groups:
              {group_and_project}:
                badges:
                  a_different_key:
                    name: "Project Badge 2"
                    link_url: "https://gitlab.example.com/foobar"
                    image_url: "https://gitlab.example.com/foobar"
                  and_also_different_than_before:
                    name: "Project Badge"
                    delete: true
            """
        run_gitlabform(config, group_and_project)

        badges = gitlab.get_project_badges(group_and_project)
        assert len(badges) == 1

        for badge in badges:
            if badge["name"] == "Project Badge 2":
                assert badge["link_url"].endswith("foobar")
                assert badge["image_url"].endswith("foobar")
            else:
                assert not badge["link_url"].endswith("foobar")
                assert not badge["image_url"].endswith("foobar")

    def test__badges_enforce(self, gitlab, group_and_project_for_function):
        group_and_project = group_and_project_for_function

        config = f"""
            projects_and_groups:
              {group_and_project}:
                badges:
                  pipeline-status:
                    name: "Project Badge"
                    link_url: "https://gitlab.example.com/first"
                    image_url: "https://gitlab.example.com/first"
                  another:
                    name: "Project Badge 2"
                    link_url: "https://gitlab.example.com/second"
                    image_url: "https://gitlab.example.com/second" 
            """
        run_gitlabform(config, group_and_project)

        badges = gitlab.get_project_badges(group_and_project)
        assert len(badges) == 2

        config2 = f"""
            projects_and_groups:
              {group_and_project}:
                badges:
                  another:
                    name: "Project Badge 2"
                    link_url: "https://gitlab.example.com/foobar"
                    image_url: "https://gitlab.example.com/foobar"
                  enforce: true
            """
        run_gitlabform(config2, group_and_project)

        badges = gitlab.get_project_badges(group_and_project)
        assert len(badges) == 1



import sys
from logging import debug

import argparse
import cli_ui
import logging
import luddite
import pkg_resources
import textwrap
import traceback
from cli_ui import (
    Symbol,
    reset,
    blue,
    message,
    error,
    info,
    fatal,
    info_1,
    debug as verbose,
    red,
    green,
    yellow,
    Token,
    purple,
    warning,
)
from packaging import version as packaging_version
from typing import Any, Tuple

from gitlabform.configuration import Configuration
from gitlabform.configuration.core import (
    ConfigFileNotFoundException,
    ConfigInvalidException,
)
from gitlabform.configuration.transform import ConfigurationTransformers
from gitlabform.constants import EXIT_INVALID_INPUT, EXIT_PROCESSING_ERROR
from gitlabform.gitlab import GitLab
from gitlabform.gitlab.core import TestRequestFailedException
from gitlabform.lists import Entities
from gitlabform.lists.filter import GroupsAndProjectsFilters
from gitlabform.lists.groups import GroupsProvider
from gitlabform.lists.projects import ProjectsProvider
from gitlabform.output import EffectiveConfigurationFile
from gitlabform.processors.group import GroupProcessors
from gitlabform.processors.project import ProjectProcessors


class GitLabForm:
    def __init__(self, include_archived_projects=True, target=None, config_string=None):

        if target and config_string:
            # this mode is basically only for testing

            self.target = target
            self.config_string = config_string
            self.verbose = True
            self.debug = True
            self.strict = True
            self.start_from = 1
            self.start_from_group = 1
            self.noop = False
            self.output_file = None
            self.skip_version_check = True
            self.include_archived_projects = include_archived_projects
            self.just_show_version = False
            self.terminate_after_error = True
            self.only_sections = "all"

            self._configure_output(tests=True)
        else:
            # normal mode

            (
                self.target,
                self.config,
                self.verbose,
                self.debug,
                self.strict,
                self.start_from,
                self.start_from_group,
                self.noop,
                self.output_file,
                self.skip_version_check,
                self.include_archived_projects,
                self.just_show_version,
                self.terminate_after_error,
                self.only_sections,
            ) = self._parse_args()

            self._configure_output()

            self._show_version(self.skip_version_check)
            if self.just_show_version:
                sys.exit(0)

            if not self.target:
                fatal(
                    "target parameter is required.",
                    exit_code=EXIT_INVALID_INPUT,
                )

        self.gitlab, self.configuration = self._initialize_configuration_and_gitlab()

        self.group_processors = GroupProcessors(
            self.gitlab, self.configuration, self.strict
        )
        self.project_processors = ProjectProcessors(
            self.gitlab, self.configuration, self.strict
        )
        self.groups_provider = GroupsProvider(
            self.gitlab,
            self.configuration,
        )
        self.projects_provider = ProjectsProvider(
            self.gitlab,
            self.configuration,
            self.include_archived_projects,
        )

        self.groups_and_projects_filters = GroupsAndProjectsFilters(
            self.configuration, self.group_processors, self.project_processors
        )

    @staticmethod
    def _parse_args() -> Tuple:
        """
        Parses the input command-line arguments.

        :return: a tuple with all the arguments that have been parsed
        """
        parser = argparse.ArgumentParser(
            description=textwrap.dedent(
                f"""
            🏗 Specialized "configuration as a code" tool for GitLab projects, groups and more
            using hierarchical configuration written in YAML.

            Exits with code:
              * 0 - on success,
              * {EXIT_INVALID_INPUT} - on invalid input errors (f.e. bad syntax in the config file) ~ "it's your fault". 😅
              * {EXIT_PROCESSING_ERROR} - if there were backend processing errors (f.e. when requests to GitLab fail) ~ "it's not your fault". 😎
            """
            ),
            formatter_class=Formatter,
        )

        parser.add_argument(
            "target",
            nargs="?",
            help='Project name in "group/project" format \n'
            "OR a single group name \n"
            'OR "ALL_DEFINED" to run for all groups and projects defined the config \n'
            'OR "ALL" to run for all projects that you have access to',
        )

        parser.add_argument(
            "-V",
            "--version",
            dest="just_show_version",
            action="store_true",
            help="show version and exit",
        )

        parser.add_argument(
            "-k",
            "--skip-version-check",
            dest="skip_version_check",
            action="store_true",
            help="Skips checking if the latest version is used",
        )

        parser.add_argument(
            "-c", "--config", default="config.yml", help="config file path and filename"
        )

        verbosity_args = parser.add_mutually_exclusive_group()

        verbosity_args.add_argument(
            "-v", "--verbose", action="store_true", help="verbose output"
        )

        verbosity_args.add_argument(
            "-d",
            "--debug",
            action="store_true",
            help="debug output (!!! WARNING !!!: sensitive data and secrets may"
            " be printed in this mode - all the data sent to GitLab API"
            " will be printed in plain-text.)",
        )

        parser.add_argument(
            "-s",
            "--strict",
            action="store_true",
            help="stop on missing branches and tags",
        )

        parser.add_argument(
            "-n",
            "--noop",
            dest="noop",
            action="store_true",
            help="run in no-op (dry run) mode",
        )

        parser.add_argument(
            "-o",
            "--output-file",
            dest="output_file",
            default=None,
            help="name/path of a file to write the effective configs to"
            " (!!! WARNING !!!: if your config contains sensitive data or secrets, then this file will also"
            " contain them.)",
        )

        parser.add_argument(
            "-a",
            "--include-archived-projects",
            dest="include_archived_projects",
            action="store_true",
            help="Includes processing projects that are archived",
        )

        parser.add_argument(
            "-t",
            "--terminate",
            dest="terminate_after_error",
            action="store_true",
            help=f"exit with {EXIT_PROCESSING_ERROR} after the first group/project processing error."
            f" (default: process all the requested groups/projects and skip the failing ones."
            f" At the end, if there were any groups/projects, exit with {EXIT_PROCESSING_ERROR}.)",
        )

        parser.add_argument(
            "-sf",
            "--start-from",
            dest="start_from",
            default=1,
            type=int,
            help="start processing projects from the given one"
            ' (as numbered by "x/y Processing group/project" messages)',
        )

        parser.add_argument(
            "-sfg",
            "--start-from-group",
            dest="start_from_group",
            default=1,
            type=int,
            help="start processing groups from the given one "
            '(as numbered by "x/y Processing group/project" messages)',
        )

        parser.add_argument(
            "-os",
            "--only-sections",
            dest="only_sections",
            default="all",
            type=str,
            help="process only section with these names (comma-delimited)",
        )

        args = parser.parse_args()

        if args.only_sections != "all":
            args.only_sections = args.only_sections.split(",")

        return (
            args.target,
            args.config,
            args.verbose,
            args.debug,
            args.strict,
            args.start_from,
            args.start_from_group,
            args.noop,
            args.output_file,
            args.skip_version_check,
            args.include_archived_projects,
            args.just_show_version,
            args.terminate_after_error,
            args.only_sections,
        )

    def _configure_output(self, tests=False) -> None:
        """
        Configures the application output using cli_ui and logging, based on debug and verbose flags:

        * normal mode - print cli_ui.* except debug as verbose
        * verbose mode - print all cli_ui.*, including debug as verbose
        * debug / tests mode - like above + (logging.)debug

        :param tests: True if we are running in tests mode
        """

        logging.basicConfig()

        if self.debug or tests:
            # debug / tests
            cli_ui_verbose = True
            level = logging.DEBUG
        elif self.verbose:
            # verbose
            cli_ui_verbose = True
            # de facto disabled as we don't use logging different from debug in this project
            level = logging.FATAL
        else:
            # normal
            cli_ui_verbose = False
            # de facto disabled as we don't use logging different from debug in this project
            level = logging.FATAL

        cli_ui.setup(verbose=cli_ui_verbose)
        logging.getLogger().setLevel(level)

        fmt = logging.Formatter("%(message)s")
        logging.getLogger().handlers[0].setFormatter(fmt)

    def _initialize_configuration_and_gitlab(self) -> Tuple[GitLab, Configuration]:
        """
        Creates the GitLab object, which represents connection to the GitLab API,
        and the configuration object, which represents the YAML configuration,
        and runs processing the configuration with configuration transformers.

        :return: tuple with GitLab and Configuration objects
        """

        try:
            if hasattr(self, "config_string"):
                gitlab = GitLab(config_string=self.config_string)
            else:
                gitlab = GitLab(config_path=self.config)
            configuration = gitlab.get_configuration()

            configuration_transformers = ConfigurationTransformers(gitlab)
            configuration_transformers.transform(configuration)

        except ConfigFileNotFoundException as e:
            fatal(
                f"Config file not found at: {e}",
                exit_code=EXIT_INVALID_INPUT,
            )
        except ConfigInvalidException as e:
            fatal(
                f"Invalid config:\n{e.underlying}",
                exit_code=EXIT_INVALID_INPUT,
            )
        except TestRequestFailedException as e:
            fatal(
                f"GitLab test request failed:\n{e.underlying}",
                exit_code=EXIT_PROCESSING_ERROR,
            )

        return gitlab, configuration

    def run(self) -> None:
        """
        The main method.
        """

        projects, groups = self._show_header(
            self.target,
        )

        group_number = 0
        successful_groups = 0
        failed_groups = {}

        effective_configuration = EffectiveConfigurationFile(self.output_file)

        for group in groups:

            group_number += 1

            if group_number < self.start_from_group:
                self._info_group_count(
                    "@",
                    group_number,
                    len(groups),
                    yellow,
                    f"Skipping group {group} as requested to start from {self.start_from_group}...",
                    reset,
                )
                continue

            group_configuration = self.configuration.get_effective_config_for_group(
                group
            )

            effective_configuration.add_placeholder(group)

            self._info_group_count(
                "@",
                group_number,
                len(groups),
                f"Processing group: {group}",
            )

            try:
                self.group_processors.process_entity(
                    group,
                    group_configuration,
                    dry_run=self.noop,
                    effective_configuration=effective_configuration,
                    only_sections=self.only_sections,
                )

                successful_groups += 1

            except Exception as e:

                failed_groups[group_number] = group

                trace = traceback.format_exc()
                message = (
                    f"Error occurred while processing group {group}, exception:\n\n{e}"
                )

                if self.terminate_after_error:
                    effective_configuration.write_to_file()
                    error(message)
                    debug(trace)
                    sys.exit(EXIT_PROCESSING_ERROR)
                else:
                    warning(message)
                    debug(trace)
            finally:
                debug(
                    f"@ ({group_number}/{len(groups)}) FINISHED Processing group: {group}"
                )

        project_number = 0
        successful_projects = 0
        failed_projects = {}

        for project_and_group in projects:

            project_number += 1

            if project_number < self.start_from:
                self._info_project_count(
                    "*",
                    project_number,
                    len(projects),
                    yellow,
                    f"Skipping project {project_and_group} as requested to start from {self.start_from}...",
                    reset,
                )
                continue

            project_configuration = self.configuration.get_effective_config_for_project(
                project_and_group
            )

            effective_configuration.add_placeholder(project_and_group)

            self._info_project_count(
                "*",
                project_number,
                len(projects),
                f"Processing project: {project_and_group}",
            )

            try:
                self.project_processors.process_entity(
                    project_and_group,
                    project_configuration,
                    dry_run=self.noop,
                    effective_configuration=effective_configuration,
                    only_sections=self.only_sections,
                )

                successful_projects += 1

            except Exception as e:

                failed_projects[project_number] = project_and_group

                trace = traceback.format_exc()
                message = f"Error occurred while processing project {project_and_group}, exception:\n\n{e}"

                if self.terminate_after_error:
                    effective_configuration.write_to_file()
                    error(message)
                    debug(trace)
                    sys.exit(EXIT_PROCESSING_ERROR)
                else:
                    warning(message)
                    debug(trace)
            finally:

                debug(
                    f"* ({project_number}/{len(projects)})"
                    f" FINISHED Processing project: {project_and_group}",
                )

        effective_configuration.write_to_file()

        self._show_summary(
            groups,
            projects,
            successful_groups,
            successful_projects,
            failed_groups,
            failed_projects,
        )

    @classmethod
    def _show_version(cls, skip_version_check: bool) -> None:
        """
        Prints the app version and how it relates to the latest stable version
        available at PyPI.

        :param skip_version_check: if True then the comparison to the latest versions is skipped
        """

        local_version = pkg_resources.get_distribution("gitlabform").version

        # fmt: off
        tower_crane = Symbol("🏗", "")
        to_show = [reset, tower_crane, "GitLabForm version:", blue, local_version, reset]
        # fmt: on
        message(*to_show, sep=" ", end="")

        if skip_version_check:
            # just print end of the line
            print()
        else:
            try:
                latest_version = luddite.get_version_pypi("gitlabform")
            except Exception as e:
                # end the line with current version
                print()
                error(f"Checking latest version failed:\n{e}")
                return

            if local_version == latest_version:
                # fmt: off
                happy = Symbol("😊", ":)")
                to_show = ["= the latest stable ", happy]
                # fmt: on
            elif packaging_version.parse(local_version) < packaging_version.parse(
                latest_version
            ):
                # fmt: off
                sad = Symbol("😔", ":(")
                to_show = ["= outdated ", sad, " , please update! (the latest stable is ", latest_version, ")"]
                # fmt: on
            else:
                # fmt: off
                excited = Symbol("🤩", "8)")
                to_show = ["= pre-release ", excited, " (the latest stable is ", latest_version, ")"]
                # fmt: on

            # complete the line with a line ending
            message(*to_show, sep="")

    def _show_header(
        self,
        target: str,
    ) -> Tuple[list, list]:
        """
        Gets the list of groups and projects to apply the configuration to, based on the provided 'target' parameter
        and prints out the output header based on that.

        :param target: what the configuration should be applied to
        :return: a tuple of lists of effective projects and effective groups
        """

        if target == "ALL":
            info(">>> Getting ALL groups and projects...")
        elif target == "ALL_DEFINED":
            info(">>> Getting all groups and projects defined in the configuration...")
        else:
            info(">>> Getting requested groups or projects...")

        groups = self.groups_provider.get_groups(target)
        projects = self.projects_provider.get_projects(target)

        if len(groups.get_effective()) == 0 and len(projects.get_effective()) == 0:
            if target == "ALL":
                error_message = "GitLab has no projects and groups!"
            elif target == "ALL_DEFINED":
                error_message = (
                    "Configuration does not have any groups or projects defined!"
                )
            else:
                error_message = f"Project or group {target} cannot be found in GitLab!"
            fatal(
                error_message,
                exit_code=EXIT_INVALID_INPUT,
            )

        self.groups_and_projects_filters.filter(groups, projects)

        self._show_input_entities(groups)
        self._show_input_entities(projects)

        return projects.get_effective(), groups.get_effective()

    @classmethod
    def _show_input_entities(cls, entities: Entities) -> None:
        """
        Prints out the groups or projects that will be processed.

        :param entities: groups or projects
        """
        info_1(f"# of {entities.name} to process: {len(entities.get_effective())}")

        entities_omitted = ""
        entities_verbose = f"{entities.name}: {entities.get_effective()}"
        if entities.any_omitted():
            entities_omitted += f"(# of omitted {entities.name} -"
            first = True
            for reason in entities.omitted:
                if len(entities.omitted[reason]) > 0:
                    if not first:
                        entities_omitted += ","
                    entities_omitted += (
                        f" {reason.value}: {len(entities.omitted[reason])}"
                    )
                    entities_verbose += f"\nomitted {entities.name} - {reason.value}: {entities.get_omitted(reason)}"
                    first = False
            entities_omitted += ")"

        if entities_omitted:
            info_1(entities_omitted)

        verbose(entities_verbose)

    @classmethod
    def _show_summary(
        cls,
        effective_groups: list,
        effective_projects: list,
        successful_groups: int,
        successful_projects: int,
        failed_groups: dict,
        failed_projects: dict,
    ):
        """
        Prints out the summary after processing has ended with the info of what was done and what failed.

        :param effective_groups: list of effective groups that the run was done on
        :param effective_projects: list of effective projects that the run was done on
        :param successful_groups: number of successfully processed groups
        :param successful_projects: number of successfully processed projects
        :param failed_groups: a dict with failed groups, where keys are their numbers in the processing order
        :param failed_projects: a dict with failed projects, where keys are their numbers in the processing order
        """

        if len(effective_groups) > 0 or len(effective_projects) > 0:
            info_1(f"# of groups processed successfully: {successful_groups}")
            info_1(f"# of projects processed successfully: {successful_projects}")

        if len(failed_groups) > 0:
            info_1(red, f"# of groups failed: {len(failed_groups)}", reset)
            for group_number in failed_groups.keys():
                # fmt: off
                info_1(red, f"Failed group {group_number}: {failed_groups[group_number]}", reset)
                # fmt: on
        if len(failed_projects) > 0:
            # fmt: off
            info_1(red, f"# of projects failed: {len(failed_projects)}", reset)
            # fmt: on
            for project_number in failed_projects.keys():
                # fmt: off
                info_1(red, f"Failed project {project_number}: {failed_projects[project_number]}", reset)
                # fmt: on

        if len(failed_groups) > 0 or len(failed_projects) > 0:
            sys.exit(EXIT_PROCESSING_ERROR)
        elif successful_groups > 0 or successful_projects > 0:
            # fmt: off
            shine = Symbol("✨", "!!!")
            info_1(green, "All requested groups/projects processed successfully!", reset, shine)
            # fmt: on
        else:
            # fmt: off
            info_1(yellow, "Nothing to do.", reset)
            # fmt: on

    @classmethod
    def _info_group_count(
        cls, prefix, i: int, n: int, *rest: Token, **kwargs: Any
    ) -> None:
        cls._info_count(purple, prefix, i, n, *rest, **kwargs)

    @classmethod
    def _info_project_count(
        cls, prefix, i: int, n: int, *rest: Token, **kwargs: Any
    ) -> None:
        cls._info_count(green, prefix, i, n, *rest, **kwargs)

    @classmethod
    def _info_count(
        cls, color, prefix, i: int, n: int, *rest: Token, **kwargs: Any
    ) -> None:
        num_digits = len(str(n))
        counter_format = f"(%{num_digits}d/%d)"
        counter_str = counter_format % (i, n)
        info(color, prefix, reset, counter_str, reset, *rest, **kwargs)


class Formatter(
    argparse.ArgumentDefaultsHelpFormatter,
    argparse.RawTextHelpFormatter,
):
    pass



from gitlabform.gitlab import GitLab
from gitlabform.processors.defining_keys import Key, And
from gitlabform.processors.multiple_entities_processor import MultipleEntitiesProcessor


class BadgesProcessor(MultipleEntitiesProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__(
            "badges",
            gitlab,
            list_method_name="get_project_badges",
            add_method_name="add_project_badge",
            delete_method_name="delete_project_badge",
            defining=Key("name"),
            required_to_create_or_update=And(
                Key("name"), Key("link_url"), Key("image_url")
            ),
            edit_method_name="edit_project_badge",
        )



import os
import textwrap

from xkcdpass import xkcd_password as xp

from gitlabform import GitLabForm
from gitlabform.gitlab import GitLab, AccessLevel
from gitlabform.gitlab.core import NotFoundException, UnexpectedResponseException

CONFIG = """
config_version: 3
"""

DEFAULT_README = "Default README content."

# automate reading files created by run_gitlab_in_docker.sh to run tests in PyCharm / IntelliJ
# (workaround for lack of this feature: https://youtrack.jetbrains.com/issue/PY-5543 )

env_vars_to_files = {
    "GITLAB_URL": "gitlab_url.txt",
    "GITLAB_TOKEN": "gitlab_token.txt",
}

for env_var in env_vars_to_files.keys():
    if env_var not in os.environ:
        print(f"{env_var} not set - trying to read it from a file...")
        for up_dir_level in range(0, 4):
            file_path = (up_dir_level * "../") + env_vars_to_files[env_var]
            print(f"Trying to read {file_path} ...")
            if os.path.isfile(file_path):
                try:
                    with open(file_path) as file:
                        os.environ[env_var] = file.read().strip()
                        print(f"{env_var} set!")
                        break
                except Exception as e:
                    print(f"Failed to read {file_path}: {e}")
            else:
                print(f"{file_path} doesn't exist.")

gl = GitLab(config_string=CONFIG)


def get_random_name(entity: str) -> str:
    random_suffix = get_random_suffix()
    return f"gitlabform_{entity}_{random_suffix}"


word_file = xp.locate_wordfile()
my_words = xp.generate_wordlist(wordfile=word_file, min_length=5, max_length=8)


def get_random_suffix():
    return xp.generate_xkcdpassword(wordlist=my_words, numwords=1, delimiter="_")


def get_random_password():
    return xp.generate_xkcdpassword(wordlist=my_words, numwords=3, delimiter="_")


def get_gitlab():
    return gl


def create_group(group_name, parent_id=None):
    gl.create_group(group_name, group_name, parent_id=parent_id, visibility="internal")


def create_groups(group_base_name, no_of_groups):
    groups = []
    for group_no in range(1, no_of_groups + 1):
        group_name = group_base_name + str(group_no)
        try:
            gl.get_group_case_insensitive(group_name)
        except NotFoundException:
            gl.create_group(group_name, group_name)
        groups.append(group_name)
    return groups


def delete_groups(group_base_name, no_of_groups):
    for group_no in range(1, no_of_groups + 1):
        group_name = group_base_name + str(group_no)
        gl.delete_group(group_name)


def create_project(group_name, project_name):
    group = gl.get_group(group_name)
    gl.create_project(
        project_name,
        project_name,
        group["id"],
        default_branch="main",
        wait_if_still_being_deleted=True,
    )

    gl.add_file(
        f"{group_name}/{project_name}",
        "main",
        "README.md",
        DEFAULT_README,
        "Create README",
    )


def create_users(user_base_name, no_of_users):
    users = []
    for user_no in range(1, no_of_users + 1):
        username = user_base_name + str(user_no)
        try:
            gl.get_user_by_name(username)
        except NotFoundException:
            gl.create_user(
                username + "@example.com",
                username + " Example",
                username,
                get_random_password(),
            )
        users.append(username)
    return users


def delete_users(user_base_name, no_of_users):
    for user_no in range(1, no_of_users + 1):
        username = user_base_name + str(user_no)
        gl.delete_user(username)


def remove_users_from_project(user_base_name, no_of_users, project_and_group):
    for user_no in range(1, no_of_users + 1):
        username = user_base_name + str(user_no)
        gl.remove_member_from_project(project_and_group, username)


def add_users_to_group(group_name, usernames, access_level=AccessLevel.DEVELOPER.value):
    for username in usernames:
        try:
            gl.add_member_to_group(group_name, username, access_level)
        except UnexpectedResponseException:
            # this is fine - user is already in the group
            pass


def remove_users_from_group(group_name, usernames):
    for username in usernames:
        try:
            gl.remove_member_from_group(group_name, username)
        except NotFoundException:
            # this is fine - user is removed from group
            pass


def delete_variables_from_group(group_name, variables):
    for variable in variables:
        try:
            gl.delete_group_variable(group_name, variable)
        except NotFoundException:
            # this is fine - variable is removed from group
            pass


def delete_pipeline_schedules_from_project(project_and_group):
    schedules = gl.get_all_pipeline_schedules(project_and_group)
    for schedule in schedules:
        gl.delete_pipeline_schedule(project_and_group, schedule["id"])


def run_gitlabform(config, target, include_archived_projects=True):
    # f-strings with """ used as configs have the disadvantage of having indentation in them - let's remove it here
    config = textwrap.dedent(config)

    # we don't want to repeat ourselves in the tests, so prefix the configs with this mandatory part here
    config = CONFIG + config

    gf = GitLabForm(
        include_archived_projects=include_archived_projects,
        config_string=config,
        target=target,
    )
    gf.run()



from logging import debug
from cli_ui import warning, fatal

from gitlabform.constants import EXIT_PROCESSING_ERROR
from gitlabform.gitlab import GitLab
from gitlabform.gitlab.core import NotFoundException
from gitlabform.processors.abstract_processor import AbstractProcessor


class TagsProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab, strict: bool):
        super().__init__("tags", gitlab)
        self.strict = strict

    def _process_configuration(self, project_and_group: str, configuration: dict):
        for tag in sorted(configuration["tags"]):
            try:
                if configuration["tags"][tag]["protected"]:
                    create_access_level = (
                        configuration["tags"][tag]["create_access_level"]
                        if "create_access_level" in configuration["tags"][tag]
                        else None
                    )
                    debug("Setting tag '%s' as *protected*", tag)
                    try:
                        # try to unprotect first
                        self.gitlab.unprotect_tag(project_and_group, tag)
                    except NotFoundException:
                        pass
                    self.gitlab.protect_tag(project_and_group, tag, create_access_level)
                else:
                    debug("Setting tag '%s' as *unprotected*", tag)
                    self.gitlab.unprotect_tag(project_and_group, tag)
            except NotFoundException:
                message = f"Tag '{tag}' not found when trying to set it as protected/unprotected!"
                if self.strict:
                    fatal(
                        message,
                        exit_code=EXIT_PROCESSING_ERROR,
                    )
                else:
                    warning(message)



from gitlabform.gitlab import GitLab
from gitlabform.processors.defining_keys import And, Or, Key, Xor
from gitlabform.processors.multiple_entities_processor import MultipleEntitiesProcessor


class GroupLDAPLinksProcessor(MultipleEntitiesProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__(
            "group_ldap_links",
            gitlab,
            list_method_name="get_ldap_group_links",
            add_method_name="add_ldap_group_link",
            delete_method_name="delete_ldap_group_link",
            defining=And(Key("provider"), Or(Key("cn"), Key("filter"))),
            required_to_create_or_update=And(
                Key("provider"), Xor(Key("cn"), Key("filter"))
            ),
        )



from cli_ui import debug as verbose

from gitlabform.gitlab.projects import GitLabProjects


class GitLabProjectProtectedEnvironments(GitLabProjects):
    def list_protected_environments(self, project_and_group_name: str):
        return self._make_requests_to_api(
            "projects/%s/protected_environments", project_and_group_name
        )

    def protect_a_repository_environment(
        self, project_and_group_name: str, protected_env_cfg: dict, retry: bool = True
    ):
        response = self._make_requests_to_api(
            "projects/%s/protected_environments",
            project_and_group_name,
            method="POST",
            json=protected_env_cfg,
            expected_codes=201,
        )

        # TODO: remove this when this issue is resolved -> https://gitlab.com/gitlab-org/gitlab/-/issues/378657
        if retry and (
            len(protected_env_cfg["deploy_access_levels"])
            != len(response["deploy_access_levels"])
        ):
            verbose(
                f'Gitlab\'s returned "deploy_access_levels" differs from the sent cfg, trying again...'
            )

            self.unprotect_environment(project_and_group_name, protected_env_cfg)

            return self.protect_a_repository_environment(
                project_and_group_name, protected_env_cfg, False
            )

        return response

    def unprotect_environment(
        self, project_and_group_name: str, protected_env_cfg: dict
    ):
        return self._make_requests_to_api(
            "projects/%s/protected_environments/%s",
            (project_and_group_name, protected_env_cfg["name"]),
            method="DELETE",
            expected_codes=204,
        )



from gitlabform.gitlab import GitLab
from gitlabform.processors.single_entity_processor import SingleEntityProcessor


class ProjectSettingsProcessor(SingleEntityProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__(
            "project_settings",
            gitlab,
            get_method_name="get_project_settings",
            edit_method_name="put_project_settings",
        )



import pytest
from tests.acceptance import run_gitlabform
import time


@pytest.fixture(scope="function")
def add_gitlab_ci_config(group_and_project):
    add_gitlab_ci_config = f"""
    projects_and_groups:
      {group_and_project}:
        files:
          ".gitlab-ci.yml":
            branches: 
              - main
            overwrite: true
            content: |
              stages:
                - deploy

              deploy:
                stage: deploy
                script:
                  - echo deploy
                environment: production
                resource_group: production
        """
    run_gitlabform(add_gitlab_ci_config, group_and_project)
    return group_and_project


class TestResourceGroups:
    def test__update_resource_group_process_mode(
        self, gitlab, group_and_project, add_gitlab_ci_config
    ):
        update_resource_group_config = f"""
        projects_and_groups:
          {group_and_project}:
            resource_groups:
              production:
                process_mode: newest_first
            """

        time.sleep(5)
        run_gitlabform(update_resource_group_config, group_and_project)

        resource_group = gitlab.get_specific_resource_group(
            group_and_project, "production"
        )
        assert resource_group["key"] == "production"
        assert resource_group["process_mode"] == "newest_first"



import pytest

from gitlabform.gitlab import AccessLevel
from tests.acceptance import (
    run_gitlabform,
)


class TestMembers:
    def test__add_user(self, gitlab, group_and_project, three_members, outsider_user):
        members_before = gitlab.get_project_members(group_and_project)
        assert len(members_before) > 0

        members_usernames_before = [member["username"] for member in members_before]
        assert outsider_user not in members_usernames_before

        add_users = f"""
        projects_and_groups:
          {group_and_project}:
            members:
              users:
                {outsider_user}: # new user
                  access_level: {AccessLevel.DEVELOPER.value}
        """

        run_gitlabform(add_users, group_and_project)

        members = gitlab.get_project_members(group_and_project)
        assert len(members) == len(members_before) + 1

        members_usernames = [member["username"] for member in members]
        assert outsider_user in members_usernames

    def test__no_groups_and_no_users(self, gitlab, group_and_project):

        config_with_error = f"""
        projects_and_groups:
          {group_and_project}:
            members:
              # there should be a sub-key 'users' here, not directly a user
              some_user1: 
                access_level: {AccessLevel.DEVELOPER.value}
        """

        with pytest.raises(SystemExit):
            run_gitlabform(config_with_error, group_and_project)

    def test__add_user_with_access_level_names(
        self, gitlab, group_and_project, three_members, outsider_user
    ):
        members_before = gitlab.get_project_members(group_and_project)
        assert len(members_before) > 0

        members_usernames_before = [member["username"] for member in members_before]
        assert outsider_user not in members_usernames_before

        add_users = f"""
            projects_and_groups:
              {group_and_project}:
                members:
                  users:
                    {outsider_user}: # new user
                      access_level: maintainer
            """

        run_gitlabform(add_users, group_and_project)

        members = gitlab.get_project_members(group_and_project)
        assert len(members) == len(members_usernames_before) + 1

        members_usernames = [member["username"] for member in members]
        assert outsider_user in members_usernames



from tests.acceptance import run_gitlabform, DEFAULT_README


class TestFilesProtected:

    # this test should be in a separate class than other test files as it changes too
    # much for a reasonable setup and cleanup using fixtures
    def test__set_file_protected_branches(
        self, gitlab, group_and_project, branch, other_branch
    ):

        set_file_protected_branches = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              main:
                protected: true
                push_access_level: maintainer
                merge_access_level: developer
                unprotect_access_level: maintainer
              {branch}:
                protected: false
              {other_branch}:
                protected: true
                push_access_level: developer
                merge_access_level: developer
                unprotect_access_level: maintainer
            files:
              "README.md":
                overwrite: true
                branches: protected
                content: "Content for protected branches only"
        """

        run_gitlabform(set_file_protected_branches, group_and_project)

        for some_branch in [
            "main",  # main branch is protected by default
            other_branch,
        ]:
            file_content = gitlab.get_file(group_and_project, some_branch, "README.md")
            assert file_content == "Content for protected branches only"
            some_branch = gitlab.get_branch(group_and_project, some_branch)
            assert some_branch["protected"] is True

        for some_branch in [branch]:
            file_content = gitlab.get_file(group_and_project, some_branch, "README.md")
            assert file_content == DEFAULT_README
            some_branch = gitlab.get_branch(group_and_project, some_branch)
            assert some_branch["protected"] is False



from gitlabform.gitlab import GitLab
from gitlabform.processors.defining_keys import Key, And
from gitlabform.processors.multiple_entities_processor import MultipleEntitiesProcessor


class GroupVariablesProcessor(MultipleEntitiesProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__(
            "group_variables",
            gitlab,
            list_method_name="get_group_variables",
            add_method_name="post_group_variable",
            delete_method_name="delete_group_variable",
            defining=Key("key"),
            required_to_create_or_update=And(Key("key"), Key("value")),
            edit_method_name="put_group_variable",
        )



import pytest

from gitlabform.gitlab import AccessLevel
from tests.acceptance import (
    run_gitlabform,
)


@pytest.fixture(scope="class")
def two_members_in_other_group(gitlab, other_group, make_user):
    outsider_user1 = make_user(add_to_project=False)
    outsider_user2 = make_user(add_to_project=False)

    gitlab.add_member_to_group(
        other_group, None, AccessLevel.OWNER.value, user_id=outsider_user1.id
    )
    gitlab.add_member_to_group(
        other_group, None, AccessLevel.DEVELOPER.value, user_id=outsider_user2.id
    )

    yield [outsider_user1.name, outsider_user2.name]


class TestMembersAddGroup:

    # this test should be in a separate class than other test files as it changes too
    # much for a reasonable setup and cleanup using fixtures
    def test__add_group(
        self,
        gitlab,
        group_and_project,
        three_members,
        other_group,
        two_members_in_other_group,
    ):
        no_of_members_before = len(gitlab.get_project_members(group_and_project))
        no_of_members_of_other_group = len(gitlab.get_group_members(other_group))

        no_of_groups_shared_before = len(
            gitlab.get_shared_with_groups(group_and_project)
        )
        assert no_of_groups_shared_before == 0

        add_group = f"""
        projects_and_groups:
          {group_and_project}:
            members:
              groups:
                {other_group}:
                  group_access: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(add_group, group_and_project)

        members = gitlab.get_project_members(group_and_project, all=True)

        assert len(members) == no_of_members_before + no_of_members_of_other_group

        for member in members:
            if member["username"] in two_members_in_other_group:
                # "group_access" is the *maximum* access level, see
                # https://docs.gitlab.com/ee/user/project/members/share_project_with_groups.html#maximum-access-level
                assert member["access_level"] <= AccessLevel.MAINTAINER.value

        no_of_groups_shared = len(gitlab.get_shared_with_groups(group_and_project))
        assert no_of_groups_shared == 1



from logging import debug
from cli_ui import debug as verbose
from cli_ui import fatal

import abc
from typing import Callable, Union, Any

from gitlabform.constants import EXIT_INVALID_INPUT
from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor
from gitlabform.processors.defining_keys import AbstractKey


class MultipleEntitiesProcessor(AbstractProcessor, metaclass=abc.ABCMeta):
    def __init__(
        self,
        configuration_name: str,
        gitlab: GitLab,
        list_method_name: Union[str, Callable[[str], list]],
        add_method_name: Union[str, Callable[[str, Any], None]],
        delete_method_name: Union[str, Callable[[str, dict], None]],
        defining: AbstractKey,
        required_to_create_or_update: AbstractKey,
        edit_method_name: Union[str, Callable[[str, dict, dict], None], None] = None,
    ):
        """
        :param configuration_name: The key in the YAML cfg file that determines the start of this Entity's section
        :param gitlab: Gitlab instance
        :param list_method_name: Retrieves the entities in question
        :param add_method_name: Creates a new entity in Gitlab
        :param delete_method_name: Removes the entity from Gitlab
        :param defining: An expression that tells which fields of the entity in GitLab are identifying it.
        :param required_to_create_or_update: Entries which (according to the API documentation) are required fields
        :param edit_method_name: Edits the existing entity in Gitlab
        """

        super().__init__(configuration_name, gitlab)
        self.list_method: Callable = (
            getattr(self.gitlab, list_method_name)
            if (isinstance(list_method_name, str))
            else list_method_name
        )
        self.add_method: Callable = (
            getattr(self.gitlab, add_method_name)
            if (isinstance(add_method_name, str))
            else add_method_name
        )
        self.delete_method: Callable = (
            getattr(self.gitlab, delete_method_name)
            if (isinstance(delete_method_name, str))
            else delete_method_name
        )
        self.defining: AbstractKey = defining
        self.required_to_create_or_update: AbstractKey = required_to_create_or_update

        if edit_method_name:
            self.edit_method: Union[Callable, None] = (
                getattr(self.gitlab, edit_method_name)
                if (isinstance(edit_method_name, str))
                else edit_method_name
            )
        else:
            self.edit_method = None

    def _process_configuration(self, project_or_group: str, configuration: dict):

        entities_in_configuration = configuration[self.configuration_name]
        if "enforce" in entities_in_configuration:
            enforce = entities_in_configuration["enforce"]
            del entities_in_configuration["enforce"]
        else:
            enforce = False

        # TODO: move/convert this to a configuration validation phase
        self._find_duplicates(project_or_group, entities_in_configuration)

        entities_in_gitlab = {}
        i = 1
        for entity_in_gitlab in self.list_method(project_or_group):
            entities_in_gitlab[str(i)] = entity_in_gitlab
            i += 1

        debug(f"{self.configuration_name} BEFORE: ^^^")

        # group entities into 3 groups:
        # a) only in gitlab,
        # b) in both configuration and gitlab (matching defining keys),
        # c) only in configuration,

        entities_only_in_gitlab = {
            entity_name: entity_config
            for (entity_name, entity_config) in entities_in_gitlab.items()
            if not self._is_in(entity_config, entities_in_configuration)
        }

        entities_in_both = {
            entity_name: entity_config
            for (entity_name, entity_config) in entities_in_configuration.items()
            if self._is_in(entity_config, entities_in_gitlab)
        }

        entities_only_in_configuration = {
            entity_name: entity_config
            for (entity_name, entity_config) in entities_in_configuration.items()
            if not self._is_in(entity_config, entities_in_gitlab)
        }

        # if "enforce", then delete a)

        if enforce:
            for entity_name, entity_config in entities_only_in_gitlab.items():
                # no need to validate if we have what's needed to delete as we got the entities from gitlab
                verbose(
                    f"Deleting entity no {entity_name} of {self.configuration_name} in {project_or_group} "
                    f"as it's not in config and enforce is set to true."
                )
                self.delete_method(project_or_group, entity_config)

        # update b), if needed (or delete them if marked as "delete")

        for entity_name, entity_config in entities_in_both.items():

            entity_in_gitlab = self._is_in(entity_config, entities_in_gitlab)
            if entity_config.get("delete", False):
                self._validate_required_to_delete(
                    project_or_group, entity_name, entity_config
                )
                verbose(
                    f"Deleting {entity_name} of {self.configuration_name} in {project_or_group}"
                )
                self.delete_method(project_or_group, entity_in_gitlab)
            elif self._needs_update(entity_in_gitlab, entity_config):
                self._validate_required_to_create_or_update(
                    project_or_group, entity_name, entity_config
                )
                if self.edit_method:
                    verbose(
                        f"Editing {entity_name} of {self.configuration_name} in {project_or_group}"
                    )
                    self.edit_method(project_or_group, entity_in_gitlab, entity_config)
                    debug(f"{self.configuration_name} AFTER: ^^^")
                else:
                    verbose(
                        f" * Recreating {entity_name} of {self.configuration_name} in {project_or_group}"
                    )
                    self.delete_method(project_or_group, entity_in_gitlab)
                    self.add_method(project_or_group, entity_config)
                    debug(f"{self.configuration_name} AFTER: ^^^")
            else:
                verbose(
                    f" * {entity_name} of {self.configuration_name} in {project_or_group} doesn't need an update."
                )

        # add c) (or do nothing if marked as "delete")

        for entity_name, entity_config in entities_only_in_configuration.items():

            self._validate_required_to_create_or_update(
                project_or_group, entity_name, entity_config
            )
            verbose(
                f" * Adding {entity_name} of {self.configuration_name} in {project_or_group}"
            )
            self.add_method(project_or_group, entity_config)
            debug(f"{self.configuration_name} AFTER: ^^^")

    def _find_duplicates(self, project_or_group: str, entities_in_configuration: dict):
        for first_key, first_value in entities_in_configuration.items():
            for second_key, second_value in entities_in_configuration.items():
                if first_key != second_key:
                    if self.defining.matches(first_value, second_value):
                        fatal(
                            f"Entities {first_key} and {second_key} in {self.configuration_name} for {project_or_group}"
                            f" are the same in terms of their defining keys: {self.defining.explain()}",
                            exit_code=EXIT_INVALID_INPUT,
                        )

    def _validate_required_to_create_or_update(
        self, project_or_group: str, entity_name: str, entity_dict: dict
    ):
        if not self.required_to_create_or_update.contains(entity_dict):
            fatal(
                f"Entity {entity_name} in {self.configuration_name} for {project_or_group}"
                f" doesn't have some of its keys required to create or update:"
                f" {self.required_to_create_or_update.explain()}",
                exit_code=EXIT_INVALID_INPUT,
            )

    def _validate_required_to_delete(
        self, project_or_group: str, entity_name: str, entity_dict: dict
    ):
        if not self.defining.contains(entity_dict):
            fatal(
                f"Entity {entity_name} in {self.configuration_name} for {project_or_group}"
                f" doesn't have some of its defining keys required to delete it: {self.defining.explain()}",
                exit_code=EXIT_INVALID_INPUT,
            )

    def _is_in(self, entity: dict, dict_of_entities: dict):
        for entity_name, entity_config in dict_of_entities.items():
            if self.defining.matches(entity, entity_config):
                return entity_config

        return False



from typing import List

from gitlabform.configuration import Configuration
from gitlabform.gitlab import GitLab
from gitlabform.processors import AbstractProcessors
from gitlabform.processors.abstract_processor import AbstractProcessor
from gitlabform.processors.project.badges_processor import BadgesProcessor
from gitlabform.processors.project.branches_processor import (
    BranchesProcessor,
)
from gitlabform.processors.project.deploy_keys_processor import (
    DeployKeysProcessor,
)
from gitlabform.processors.project.files_processor import FilesProcessor
from gitlabform.processors.project.hooks_processor import HooksProcessor
from gitlabform.processors.project.members_processor import MembersProcessor
from gitlabform.processors.project.project_processor import ProjectProcessor
from gitlabform.processors.project.project_push_rules_processor import (
    ProjectPushRulesProcessor,
)
from gitlabform.processors.project.project_settings_processor import (
    ProjectSettingsProcessor,
)
from gitlabform.processors.shared.protected_environments_processor import (
    ProtectedEnvironmentsProcessor,
)
from gitlabform.processors.project.resource_groups_processor import (
    ResourceGroupsProcessor,
)
from gitlabform.processors.project.schedules_processor import (
    SchedulesProcessor,
)
from gitlabform.processors.project.variables_processor import (
    VariablesProcessor,
)
from gitlabform.processors.project.integrations_processor import (
    IntegrationsProcessor,
)
from gitlabform.processors.project.tags_processor import TagsProcessor
from gitlabform.processors.project.merge_requests_approval_rules import (
    MergeRequestsApprovalRules,
)
from gitlabform.processors.project.merge_requests_approvals import (
    MergeRequestsApprovals,
)


class ProjectProcessors(AbstractProcessors):
    def __init__(self, gitlab: GitLab, config: Configuration, strict: bool):
        super().__init__(gitlab, config, strict)
        self.processors: List[AbstractProcessor] = [
            ProjectProcessor(gitlab),
            ProjectSettingsProcessor(gitlab),
            ProjectPushRulesProcessor(gitlab),
            DeployKeysProcessor(gitlab),
            VariablesProcessor(gitlab),
            BranchesProcessor(gitlab, strict),
            TagsProcessor(gitlab, strict),
            IntegrationsProcessor(gitlab),
            FilesProcessor(gitlab, config, strict),
            HooksProcessor(gitlab),
            MembersProcessor(gitlab),
            SchedulesProcessor(gitlab),
            BadgesProcessor(gitlab),
            ResourceGroupsProcessor(gitlab),
            ProtectedEnvironmentsProcessor(gitlab),
            MergeRequestsApprovals(gitlab),
            MergeRequestsApprovalRules(gitlab),
        ]



import os

from tests.acceptance import (
    run_gitlabform,
)


class TestTokenFromConfig:
    def test__token_no_quotes(self, gitlab, group_and_project, token_from_env_var):

        config = f"""
        gitlab:
          token: {token_from_env_var}
          
        projects_and_groups:
          placeholder:
        """

        run_gitlabform(config, group_and_project)

    def test__token_single_quotes(self, gitlab, group_and_project, token_from_env_var):

        config = f"""
        gitlab:
          token: '{token_from_env_var}'
          
        projects_and_groups:
          placeholder:
        """

        run_gitlabform(config, group_and_project)

    def test__token_double_quotes(self, gitlab, group_and_project, token_from_env_var):

        config = f"""
        gitlab:
          token: "{token_from_env_var}"

        projects_and_groups:
          placeholder:
        """

        run_gitlabform(config, group_and_project)



import os
import re
from pathlib import Path

from logging import debug
from cli_ui import debug as verbose
from cli_ui import warning, fatal

from jinja2 import Environment, FileSystemLoader

from gitlabform.constants import EXIT_INVALID_INPUT
from gitlabform.configuration import Configuration
from gitlabform.gitlab import GitLab
from gitlabform.gitlab.core import NotFoundException, UnexpectedResponseException
from gitlabform.processors.abstract_processor import AbstractProcessor
from gitlabform.processors.util.branch_protector import BranchProtector


class FilesProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab, config: Configuration, strict: bool):
        super().__init__("files", gitlab)
        self.config = config
        self.strict = strict
        self.branch_protector = BranchProtector(gitlab, strict)

    def _process_configuration(self, project_and_group: str, configuration: dict):
        for file in sorted(configuration["files"]):
            debug("Processing file '%s'...", file)

            if configuration.get("files|" + file + "|skip"):
                debug("Skipping file '%s'", file)
                continue

            if configuration["files"][file]["branches"] == "all":
                all_branches = self.gitlab.get_branches(project_and_group)
                branches = sorted(all_branches)
            elif configuration["files"][file]["branches"] == "protected":
                protected_branches = self.gitlab.get_protected_branches(
                    project_and_group
                )
                branches = sorted(protected_branches)
            else:
                all_branches = self.gitlab.get_branches(project_and_group)
                branches = []
                for branch in configuration["files"][file]["branches"]:
                    if branch in all_branches:
                        branches.append(branch)
                    else:
                        message = f"! Branch '{branch}' not found, not processing file '{file}' in it"
                        if self.strict:
                            fatal(
                                message,
                                exit_code=EXIT_INVALID_INPUT,
                            )
                        else:
                            warning(message)

            for branch in branches:
                verbose(f"Processing file '{file}' in branch '{branch}'")

                if configuration.get(
                    "files|" + file + "|content"
                ) and configuration.get("files|" + file + "|file"):
                    fatal(
                        f"File '{file}' in '{project_and_group}' has both `content` and `file` set - "
                        "use only one of these keys.",
                        exit_code=EXIT_INVALID_INPUT,
                    )

                if configuration.get("files|" + file + "|delete"):
                    try:
                        self.gitlab.get_file(project_and_group, branch, file)
                        debug("Deleting file '%s' in branch '%s'", file, branch)
                        self.modify_file_dealing_with_branch_protection(
                            project_and_group,
                            branch,
                            file,
                            "delete",
                            configuration,
                        )
                    except NotFoundException:
                        debug(
                            "Not deleting file '%s' in branch '%s' (already doesn't exist)",
                            file,
                            branch,
                        )
                else:
                    # change or create file

                    if configuration.get("files|" + file + "|content"):
                        new_content = configuration.get("files|" + file + "|content")
                    else:
                        path_in_config = Path(
                            str(configuration.get("files|" + file + "|file"))
                        )
                        if path_in_config.is_absolute():
                            effective_path = path_in_config
                        else:
                            # relative paths are relative to config file location
                            effective_path = Path(
                                os.path.join(
                                    self.config.config_dir, str(path_in_config)
                                )
                            )
                        new_content = effective_path.read_text()

                    # templating is documented to be enabled by default,
                    # see https://gitlabform.github.io/gitlabform/reference/files/#files
                    templating_enabled = True

                    if configuration.get(
                        "files|" + file + "|template", templating_enabled
                    ):
                        new_content = self.get_file_content_as_template(
                            new_content,
                            project_and_group,
                            **configuration.get("files|" + file + "|jinja_env", dict()),
                        )

                    try:
                        current_content = self.gitlab.get_file(
                            project_and_group, branch, file
                        )
                        if current_content != new_content:
                            if configuration.get("files|" + file + "|overwrite"):
                                debug("Changing file '%s' in branch '%s'", file, branch)
                                self.modify_file_dealing_with_branch_protection(
                                    project_and_group,
                                    branch,
                                    file,
                                    "modify",
                                    configuration,
                                    new_content,
                                )
                            else:
                                debug(
                                    "Not changing file '%s' in branch '%s' - overwrite flag not set.",
                                    file,
                                    branch,
                                )
                        else:
                            debug(
                                "Not changing file '%s' in branch '%s' - it's content is already"
                                " as provided)",
                                file,
                                branch,
                            )
                    except NotFoundException:
                        debug("Creating file '%s' in branch '%s'", file, branch)
                        self.modify_file_dealing_with_branch_protection(
                            project_and_group,
                            branch,
                            file,
                            "add",
                            configuration,
                            new_content,
                        )

                if configuration.get("files|" + file + "|only_first_branch"):
                    verbose("Skipping other branches for this file, as configured.")
                    break

    def modify_file_dealing_with_branch_protection(
        self,
        project_and_group,
        branch,
        file,
        operation,
        configuration,
        new_content=None,
    ):
        # perhaps your user permissions are ok to just perform this operation regardless
        # of the branch protection...

        try:

            self.just_modify_file(
                project_and_group, branch, file, operation, configuration, new_content
            )

        except UnexpectedResponseException as e:

            if (
                e.response_status_code == 400
                and "You are not allowed to push into this branch" in e.response_text
            ):

                # ...but if not, then we can unprotect the branch, but only if we know how to
                # protect it again...

                if configuration.get("branches|" + branch + "|protected"):
                    debug(
                        f"> Temporarily unprotecting the branch to {operation} a file in it..."
                    )
                    self.branch_protector.unprotect_branch(project_and_group, branch)
                else:
                    fatal(
                        f"Operation {operation} on file {file} in branch {branch} not permitted,"
                        f" but we don't have a branch protection configuration provided for this"
                        f" branch. Breaking as we cannot unprotect the branch as we would not know"
                        f" how to protect it again.",
                        EXIT_INVALID_INPUT,
                    )

                try:
                    self.just_modify_file(
                        project_and_group,
                        branch,
                        file,
                        operation,
                        configuration,
                        new_content,
                    )

                finally:
                    # ...and protect the branch again after the operation
                    if configuration.get("branches|" + branch + "|protected"):
                        debug("> Protecting the branch again.")
                        self.branch_protector.protect_branch(
                            project_and_group, configuration, branch
                        )

            else:
                raise e

    def just_modify_file(
        self,
        project_and_group,
        branch,
        file,
        operation,
        configuration,
        new_content=None,
    ):
        if operation == "modify":
            self.gitlab.set_file(
                project_and_group,
                branch,
                file,
                new_content,
                self.get_commit_message_for_file_change("change", file, configuration),
            )
        elif operation == "add":
            self.gitlab.add_file(
                project_and_group,
                branch,
                file,
                new_content,
                self.get_commit_message_for_file_change("add", file, configuration),
            )
        elif operation == "delete":
            self.gitlab.delete_file(
                project_and_group,
                branch,
                file,
                self.get_commit_message_for_file_change("delete", file, configuration),
            )

    def get_file_content_as_template(self, template, project_and_group, **kwargs):
        # Use jinja with variables project and group
        rtemplate = Environment(
            loader=FileSystemLoader("."),
            autoescape=True,
            keep_trailing_newline=True,
        ).from_string(template)
        return rtemplate.render(
            project=self.get_project(project_and_group),
            group=self.get_group(project_and_group),
            **kwargs,
        )

    @staticmethod
    def get_commit_message_for_file_change(operation, file, configuration: dict):
        commit_message = configuration.get(
            "files|" + file + "|commit_message",
            "Automated %s made by gitlabform" % operation,
        )

        # add '[skip ci]' to commit message to skip CI job, as documented at
        # https://docs.gitlab.com/ee/ci/yaml/README.html#skipping-jobs
        skip_build = configuration.get("files|" + file + "|skip_ci")
        skip_build_str = " [skip ci]" if skip_build else ""

        return f"{commit_message}{skip_build_str}"

    @staticmethod
    def get_group(project_and_group):
        return re.match("(.*)/.*", project_and_group).group(1)

    @staticmethod
    def get_project(project_and_group):
        return re.match(".*/(.*)", project_and_group).group(1)



from gitlabform.gitlab.core import GitLabCore


class GitLabMergeRequests(GitLabCore):
    def create_mr(
        self,
        project_and_group_name,
        source_branch,
        target_branch,
        title,
        description=None,
    ):
        data = {
            "id": project_and_group_name,
            "source_branch": source_branch,
            "target_branch": target_branch,
            "title": title,
            "description": description,
        }
        return self._make_requests_to_api(
            "projects/%s/merge_requests",
            project_and_group_name,
            method="POST",
            data=data,
            expected_codes=201,
        )

    def accept_mr(self, project_and_group_name, mr_iid):
        return self._make_requests_to_api(
            "projects/%s/merge_requests/%s/merge",
            (project_and_group_name, mr_iid),
            method="PUT",
        )

    def update_mr(self, project_and_group_name, mr_iid, data):
        self._make_requests_to_api(
            "projects/%s/merge_requests/%s",
            (project_and_group_name, mr_iid),
            method="PUT",
            data=data,
        )

    def get_mrs(self, project_and_group_name):
        """
        :param project_and_group_name: like 'group/project'
        :return: get all *open* MRs in given project
        """
        return self._make_requests_to_api(
            "projects/%s/merge_requests?scope=all&state=opened",
            project_and_group_name,
        )

    def get_mr(self, project_and_group_name, mr_iid):
        return self._make_requests_to_api(
            "projects/%s/merge_requests/%s", (project_and_group_name, mr_iid)
        )

    def get_mr_approvals(self, project_and_group_name, mr_iid):
        return self._make_requests_to_api(
            "projects/%s/merge_requests/%s/approvals", (project_and_group_name, mr_iid)
        )



from typing import Any

import os
import logging
import textwrap
from abc import ABC
from copy import deepcopy
from logging import debug
from pathlib import Path
from ruamel.yaml.scalarstring import ScalarString
from types import SimpleNamespace

from cli_ui import debug as verbose
from cli_ui import fatal
from mergedeep import merge
from yamlpath.common import Parsers
from yamlpath.wrappers import ConsolePrinter

from gitlabform.constants import EXIT_INVALID_INPUT
from ruamel.yaml.comments import CommentedMap


class ConfigurationCore(ABC):
    """
    The functionality shared by various types of config (common, groups, projects)
    is implemented here.
    """

    def __init__(self, config_path=None, config_string=None):

        if config_path and config_string:
            fatal(
                "Please initialize with either config_path or config_string, not both.",
                exit_code=EXIT_INVALID_INPUT,
            )
        try:
            if config_string:
                self.config = self._parse_yaml(config_string, config_string=True)
                self.config_dir = "."
            else:  # maybe config_path
                config_path = self._get_config_path(config_path)
                self.config = self._parse_yaml(config_path, config_string=False)
                self.config_dir = os.path.dirname(config_path)

                # below checks are only needed in the non-test mode, when the config is read from file

                if self.config.get("example_config"):
                    fatal(
                        "Example config detected, aborting.\n"
                        "Haven't you forgotten to use `-c <config_file>` parameter?\n"
                        "If you created your config based on the example config.yml,"
                        " then please remove 'example_config' key.",
                        exit_code=EXIT_INVALID_INPUT,
                    )

                if self.config.get("config_version", 1) != 3:
                    fatal(
                        "This version of GitLabForm requires 'config_version: 3' entry in the config. "
                        "This ensures that if the application behavior changes in a backward-incompatible way,"
                        " you won't apply unwanted configuration to your GitLab instance.\n"
                        "Please follow this guide: https://gitlabform.github.io/gitlabform/upgrade/\n",
                        exit_code=EXIT_INVALID_INPUT,
                    )

            self._find_almost_duplicates()

            # we are NOT checking for the existence of non-empty 'projects_and_groups' key here
            # as it would break using GitLabForm as a library

        except (FileNotFoundError, OSError):
            raise ConfigFileNotFoundException(config_path)

        except Exception as e:
            raise ConfigInvalidException(e)

    @staticmethod
    def _get_config_path(config_path):

        if "APP_HOME" in os.environ:
            # using this env var should be considered unofficial, we need this temporarily
            # for backwards compatibility. support for it may be removed without notice, do not use it!
            config_path = os.path.join(os.environ["APP_HOME"], "config.yml")
        elif not config_path:
            # this case is only meant for using gitlabform as a library
            config_path = os.path.join(str(Path.home()), ".gitlabform", "config.yml")
        elif config_path in [os.path.join(".", "config.yml"), "config.yml"]:
            # provided points to config.yml in the app current working dir
            config_path = os.path.join(os.getcwd(), "config.yml")

        return config_path

    @staticmethod
    def _parse_yaml(source: str, config_string: bool):

        logging_args = SimpleNamespace(quiet=False, verbose=False, debug=False)
        log = ConsolePrinter(logging_args)

        yaml = Parsers.get_yaml_editor()

        # for better backward compatibility with PyYAML (that supports only YAML 1.1) used in the previous
        # GitLabForm versions, let's force ruamel.yaml to use YAML version 1.1 by default too
        yaml.version = (1, 1)

        if config_string:
            config = textwrap.dedent(source)
            verbose("Reading config from the provided string.")
            (yaml_data, doc_loaded) = Parsers.get_yaml_data(
                yaml, log, config, literal=True
            )
        else:
            config_path = source
            verbose(f"Reading config from file: {config_path}")
            (yaml_data, doc_loaded) = Parsers.get_yaml_data(yaml, log, config_path)

        if doc_loaded:
            debug("Config parsed successfully as YAML.")
        else:
            # an error message has already been printed via ConsolePrinter
            exit(EXIT_INVALID_INPUT)

        return yaml_data

    def get(self, path, default=None) -> Any:
        """
        :param path: "path" to given element in YAML file, for example for:

        group_settings:
          sddc:
            deploy_keys:
              qa_puppet:
                key: some key...
                title: some title...
                can_push: false

        ..a path to a single element array ['qa_puppet'] will be: "group_settings|sddc|deploy_keys".

        To get the dict under it use: get("group_settings|sddc|deploy_keys")

        :param default: the value to return if the key is not found. The default 'None' means that an exception
                        will be raised in such case.
        :return: element from YAML file (dict, array, string...)
        """
        tokens = path.split("|")
        current = self.config

        try:
            for token in tokens:
                current = current[token]

            if isinstance(current, ScalarString):
                to_return = str(current)
            else:
                to_return = current
        except:
            if default is not None:
                to_return = default
            else:
                raise KeyNotFoundException(path) from None

        return to_return

    @staticmethod
    def _validate_break_inheritance_flag(
        config: dict, section_name: str, parent_key: str = ""
    ) -> None:
        for key, value in config.items():
            if "inherit" == key:
                parent_key_description = (
                    ' under key "' + parent_key + '"' if parent_key else ""
                )
                fatal(
                    f'The inheritance-break flag set in "{section_name}"{parent_key_description} is invalid\n'
                    f"because it has no higher level setting to inherit from.\n",
                    exit_code=EXIT_INVALID_INPUT,
                )
            elif type(value) is CommentedMap:
                ConfigurationCore._validate_break_inheritance_flag(
                    value, section_name, key
                )

    @staticmethod
    def _merge_configs(more_general_config, more_specific_config) -> dict:
        """
        :return: merge more general config with more specific configs.
                 More specific config values take precedence over more general ones.
        """

        more_general_config = deepcopy(more_general_config)
        more_specific_config = deepcopy(more_specific_config)

        merged_dict = merge({}, more_general_config, more_specific_config)

        def break_inheritance(specific_config, parent_key=""):
            for key, value in specific_config.items():
                if "inherit" == key:
                    if not value:
                        replace_config_sections(
                            merged_dict, parent_key, specific_config
                        )
                        break
                    elif value:
                        fatal(
                            f"Cannot set the inheritance break flag with true\n",
                            exit_code=EXIT_INVALID_INPUT,
                        )
                elif type(value) is CommentedMap:
                    break_inheritance(value, key)

        def replace_config_sections(merged_config, specific_key, specific_config):
            for key, value in merged_config.items():
                if specific_key == key:
                    del specific_config["inherit"]
                    merged_config[key] = specific_config
                    break
                elif type(value) is CommentedMap:
                    replace_config_sections(value, specific_key, specific_config)

        break_inheritance(more_specific_config)

        return dict(merged_dict)

    @staticmethod
    def _get_case_insensitively(a_dict: dict, a_key: str):
        for dict_key in a_dict.keys():
            if dict_key.lower() == a_key.lower():
                return a_dict[dict_key]
        return {}

    @staticmethod
    def _is_skipped_case_insensitively(an_array: list, item: str) -> bool:
        """
        :return: if item is defined in the list to be skipped
        """
        item = item.lower()

        for list_element in an_array:
            list_element = list_element.lower()

            if list_element == item:
                return True

            if (
                list_element.endswith("/*")
                and item.startswith(list_element[:-2])
                and len(item) >= len(list_element[:-2])
            ):
                return True

        return False

    def _find_almost_duplicates(self):

        # in GitLab groups and projects names are de facto case insensitive:
        # you can change the case of both name and path BUT you cannot create
        # 2 groups which names differ only with case and the same thing for
        # projects. therefore we cannot allow such entries in the config,
        # as they would be ambiguous.

        for path in [
            "projects_and_groups",
            "skip_groups",
            "skip_projects",
        ]:
            if self.get(path, 0):
                almost_duplicates = self._find_almost_duplicates_in(path)
                if almost_duplicates:
                    fatal(
                        f"There are almost duplicates in the keys of {path} - they differ only in case.\n"
                        f"They are: {', '.join(almost_duplicates)}\n"
                        f"This is not allowed as we ignore the case for group and project names.",
                        exit_code=EXIT_INVALID_INPUT,
                    )

    def _find_almost_duplicates_in(self, configuration_path):
        """
        Checks given configuration key and reads its keys - if it is a dict - or elements - if it is a list.
        Looks for items that are almost the same - they differ only in the case.
        :param configuration_path: configuration path, f.e. "group_settings"
        :return: list of items that have almost duplicates,
                 or an empty list if none are found
        """

        dict_or_list = self.get(configuration_path)
        if isinstance(dict_or_list, dict):
            items = dict_or_list.keys()
        else:
            items = dict_or_list

        items_with_lowercase_names = [x.lower() for x in items]

        # casting these to sets will deduplicate the one with lowercase names
        # lowering its cardinality if there were elements in it
        # that before lowering differed only in case
        if len(set(items)) != len(set(items_with_lowercase_names)):

            # we have some almost duplicates, let's find them
            almost_duplicates = []
            for first_item in items:
                occurrences = 0
                for second_item in items_with_lowercase_names:
                    if first_item.lower() == second_item.lower():
                        occurrences += 1
                        if occurrences == 2:
                            almost_duplicates.append(first_item)
                            break
            return almost_duplicates

        else:
            return []


class ConfigFileNotFoundException(Exception):
    pass


class ConfigInvalidException(Exception):
    def __init__(self, underlying: Exception):
        self.underlying = underlying


class KeyNotFoundException(Exception):
    __slots__ = "key"

    def __init__(self, key: str):
        if logging.getLogger().getEffectiveLevel() <= logging.DEBUG:
            self.key = key
        else:
            fatal(
                f"Unable to find the key: {key.replace('|', '.')}\n",
                exit_code=EXIT_INVALID_INPUT,
            )



from gitlabform.processors import AbstractProcessor


class TestRecursiveDiffAnalyzer:
    _cfg_a = [
        {
            "access_level": 40,
            "access_level_description": "Maintainers",
            "user_id": None,
            "group_id": None,
            "group_inheritance_type": 0,
        },
        {
            "access_level": 40,
            "access_level_description": "John Doe",
            "user_id": 967,
            "group_id": None,
            "group_inheritance_type": 0,
        },
    ]

    _cfg_b = [
        {"access_level": 40, "group_inheritance_type": 0},
        {"user_id": 967},
    ]

    def test__equal_configurations(self) -> None:
        assert not AbstractProcessor.recursive_diff_analyzer(
            "deploy_access_levels", self._cfg_a, self._cfg_b
        )

    def test__unequal_configurations(self) -> None:
        modified_cfg = self._cfg_b.copy()

        modified_cfg[1]["group_inheritance_type"] = 1

        assert AbstractProcessor.recursive_diff_analyzer(
            "deploy_access_levels", self._cfg_a, modified_cfg
        )



from gitlabform.gitlab import GitLab
from gitlabform.processors.defining_keys import Key, And
from gitlabform.processors.multiple_entities_processor import MultipleEntitiesProcessor


class MergeRequestsApprovalRules(MultipleEntitiesProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__(
            "merge_requests_approval_rules",
            gitlab,
            list_method_name="get_approval_rules",
            add_method_name="add_approval_rule",
            edit_method_name="edit_approval_rule",
            delete_method_name="delete_approval_rule",
            defining=Key("name"),
            required_to_create_or_update=And(Key("name"), Key("approvals_required")),
        )



import enum

import copy
from collections import defaultdict

from typing import DefaultDict

from abc import ABC

from enum import Enum


@enum.unique
class OmissionReason(Enum):
    ARCHIVED = "archived"
    EMPTY = "empty effective config"
    SKIPPED = "skipped"


class Entities(ABC):
    """
    Entities here are lists of GitLab groups or projects on which GitLabForm will operate.
    The code here assumes that there is a set of requested entities, but some of them may be omitted
    for various reasons, so we want to be able to get only the effective set of them.
    """

    def __init__(self, name: str):
        self.requested: set = set()
        self.omitted: DefaultDict[OmissionReason, set] = defaultdict(set)
        self.name = name

    def add_requested(self, more_requested: list) -> None:
        self.requested = self.requested | set(more_requested)

    def add_omitted(self, reason: OmissionReason, more_omitted: list) -> None:
        self.omitted[reason] = self.omitted[reason] | set(more_omitted)

    def get_omitted(self, reason: OmissionReason) -> list:
        return sorted(self.omitted[reason])

    def any_omitted(self) -> bool:
        for reason in self.omitted:
            if len(self.omitted[reason]) > 0:
                return True
        return False

    def get_effective(self) -> list:
        effective_set = copy.deepcopy(self.requested)
        for reason in self.omitted:
            effective_set -= self.omitted[reason]
        return sorted(effective_set)


class Groups(Entities):
    def __init__(self):
        super().__init__("groups")


class Projects(Entities):
    def __init__(self):
        super().__init__("projects")



from logging import debug

import ez_yaml
from cli_ui import debug as verbose
from cli_ui import fatal

from gitlabform.constants import EXIT_INVALID_INPUT, EXIT_PROCESSING_ERROR


class EffectiveConfigurationFile:
    """
    For GitLabForm upgrades and configuration refactoring we want to be able to compare the effective configurations
    before and after the code/app change. This class provides a feature to write the effective configuration into a YAML
    file.
    """

    def __init__(self, output_file):
        if output_file:
            try:
                self.output_file = open(output_file, "w")
                debug(
                    f"Opened file {self.output_file} to write the effective configs to."
                )
            except Exception as e:
                fatal(
                    f"Error when trying to open {self.output_file} to write the effective configs to: {e}",
                    exit_code=EXIT_INVALID_INPUT,
                )
        else:
            self.output_file = None

        self.config = {}

    def add_placeholder(self, project_or_group: str):
        if self.output_file:
            self.config[project_or_group] = {}

    def add_configuration(
        self, project_or_group: str, configuration_name: str, configuration: dict
    ):
        if self.output_file:
            verbose(f"Adding effective configuration for {configuration_name}.")
            self.config[project_or_group][configuration_name] = configuration

    def write_to_file(self):
        if self.output_file:
            try:
                yaml_configuration = ez_yaml.to_string(self.config)
                self.output_file.write(yaml_configuration)
                self.output_file.close()
            except Exception as e:
                fatal(
                    f"Error when trying to write or close {self.output_file}: {e}",
                    exit_code=EXIT_PROCESSING_ERROR,
                )



from gitlabform.gitlab.core import (
    NotFoundException,
    GitLabCore,
)


class GitLabProjectMergeRequestsApprovals(GitLabCore):

    # configuration

    def get_approvals_settings(self, project_and_group_name):
        pid = self._get_project_id(project_and_group_name)
        return self._make_requests_to_api("projects/%s/approvals", pid)

    def post_approvals_settings(self, project_and_group_name, data):
        # for this endpoint GitLab still actually wants pid, not "group/project"...
        pid = self._get_project_id(project_and_group_name)
        data_required = {"id": pid}
        data = {**data, **data_required}
        self._make_requests_to_api(
            "projects/%s/approvals", pid, "POST", data, expected_codes=201
        )

    # approval rules

    def get_approval_rules(self, project_and_group_name):
        # for this endpoint GitLab still actually wants pid, not "group/project"...
        pid = self._get_project_id(project_and_group_name)
        return self._make_requests_to_api("projects/%s/approval_rules", pid)

    def get_approval_rule(self, project_and_group_name, name):
        # for this endpoint GitLab still actually wants pid, not "group/project"...
        pid = self._get_project_id(project_and_group_name)
        rules = self._make_requests_to_api("projects/%s/approval_rules", pid)
        for rule in rules:
            if rule["name"] == name:
                return rule
        raise NotFoundException

    def add_approval_rule(
        self,
        project_and_group_name,
        data,
    ):
        pid = self._get_project_id(project_and_group_name)

        if "protected_branches" in data:
            self._transform_protected_branches(data, project_and_group_name)

        self._make_requests_to_api(
            "projects/%s/approval_rules",
            pid,
            method="POST",
            data=None,
            expected_codes=201,
            json=data,
        )

    def edit_approval_rule(
        self,
        project_and_group_name,
        rule_in_gitlab,
        rule_in_config,
    ):
        pid = self._get_project_id(project_and_group_name)
        approval_rule_id = rule_in_gitlab["id"]

        # GitLab interprets not passing any of these lists as "do not change them"
        # while what we really what is in this case is "clear them"
        if "user_ids" not in rule_in_config:
            rule_in_config["user_ids"] = []
        if "group_ids" not in rule_in_config:
            rule_in_config["group_ids"] = []
        if "protected_branches" in rule_in_config:
            self._transform_protected_branches(rule_in_config, project_and_group_name)
        else:
            rule_in_config["protected_branch_ids"] = []

        self._make_requests_to_api(
            "projects/%s/approval_rules/%s",
            (pid, approval_rule_id),
            method="PUT",
            data=None,
            json=rule_in_config,
        )

    def delete_approval_rule(self, project_and_group_name, rule_in_gitlab):
        # for this endpoint GitLab still actually wants pid, not "group/project"...
        pid = self._get_project_id(project_and_group_name)
        approval_rule_id = rule_in_gitlab["id"]

        self._make_requests_to_api(
            "projects/%s/approval_rules/%s",
            (pid, approval_rule_id),
            method="DELETE",
            expected_codes=[200, 204],
        )

    def _transform_protected_branches(self, data, project_and_group_name):
        # we do this transformation here instead of in a ConfigurationTransformers
        # because we need a context of the project and group to convert branch names to ids
        # and that is not available there

        protected_branches_name = data["protected_branches"]
        data.pop("protected_branches")
        protected_branch_ids = []
        for branch_name in protected_branches_name:
            branch_id = self._get_protected_branch_id(
                project_and_group_name, branch_name
            )
            protected_branch_ids.append(branch_id)
        data["protected_branch_ids"] = protected_branch_ids



from cli_ui import fatal

from gitlabform.gitlab import GitLab
from gitlabform.processors.single_entity_processor import SingleEntityProcessor


class MergeRequestsApprovals(SingleEntityProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__(
            "merge_requests_approvals",
            gitlab,
            get_method_name="get_approvals_settings",
            edit_method_name="post_approvals_settings",
        )

    def _can_proceed(self, project_or_group: str, configuration: dict):
        if "approvals_before_merge" in configuration["merge_requests_approvals"]:
            fatal(
                "Setting the 'approvals_before_merge' in the 'merge_requests_approvals' sections is not allowed "
                "as it is not clear which rule does it apply to. "
                "Please set it inside the specific approval rules in the 'merge_requests_approval_rules' section."
            )
        else:
            return True



import os
import pytest

from tests.acceptance import run_gitlabform


@pytest.fixture(scope="class")
def file3(request):
    f = open("file3.txt", "a")
    f.write("{{ group }}/{{ project }}\n\n")
    f.close()

    def fin():
        os.remove("file3.txt")

    request.addfinalizer(fin)


class TestFilesTemplates:
    def test__default_variables(self, gitlab, group_and_project):

        config = f"""
        projects_and_groups:
          {group_and_project}:
            files:
              "README.md":
                overwrite: true
                branches: all
                content: |
                  This is a text that contains the default variables: {{{{ group }}}}/{{{{ project }}}}
        """

        run_gitlabform(config, group_and_project)

        file_content = gitlab.get_file(group_and_project, "main", "README.md")
        assert (
            file_content
            == f"This is a text that contains the default variables: {group_and_project}\n"
        )

    def test__custom_variables(self, gitlab, group_and_project):

        config = f"""
        projects_and_groups:
          {group_and_project}:
            files:
              "README.md":
                overwrite: true
                branches: all
                jinja_env:
                  foo: "fooz"
                  bar: "barz"
                content: |
                  This is a text that contains custom variables: {{{{ foo }}}}, {{{{ bar }}}}
        """

        run_gitlabform(config, group_and_project)

        file_content = gitlab.get_file(group_and_project, "main", "README.md")
        assert (
            file_content
            == "This is a text that contains custom variables: fooz, barz\n"
        )

    def test__trailing_new_lines(self, gitlab, group_and_project, file3):
        set_file_chinese_characters = f"""
        projects_and_groups:
          {group_and_project}:
            files:
              "README.md":
                overwrite: true
                branches: all
                file: file3.txt
        """

        run_gitlabform(set_file_chinese_characters, group_and_project)

        file_content = gitlab.get_file(group_and_project, "main", "README.md")
        assert file_content == f"{group_and_project}\n\n"

    def test__disabled_templating(self, gitlab, group_and_project):

        config = f"""
        projects_and_groups:
          {group_and_project}:
            files:
              "README.md":
                overwrite: true
                branches: all
                template: no
                content: |
                  This is a text containing literals: {{{{ group }}}}/{{{{ project }}}}
        """

        run_gitlabform(config, group_and_project)

        file_content = gitlab.get_file(group_and_project, "main", "README.md")
        assert (
            file_content
            == "This is a text containing literals: {{ group }}/{{ project }}\n"
        )



from gitlabform.gitlab.core import NotFoundException, InvalidParametersException
from gitlabform.gitlab.groups import GitLabGroups


class GitLabGroupLDAPLinks(GitLabGroups):
    def get_ldap_group_links(self, group):
        group_id = self.get_group_id_case_insensitive(group)
        return self._make_requests_to_api(
            "groups/%s/ldap_group_links", group_id, expected_codes=[200, 404]
        )

    def add_ldap_group_link(self, group, data):
        group_id = self.get_group_id_case_insensitive(group)
        data["id"] = group_id

        try:
            return self._make_requests_to_api(
                "groups/%s/ldap_group_links",
                group_id,
                method="POST",
                data=data,
                expected_codes=[200, 201],
            )
        # this is a GitLab API bug - it returns 404 here instead of 400 for bad requests...
        except NotFoundException:
            raise InvalidParametersException(
                f"Invalid parameters for a Group LDAP link for group {group}: {data}"
            )

    def delete_ldap_group_link(self, group, data):
        if "group_access" in data:
            del data["group_access"]

        group_id = self.get_group_id_case_insensitive(group)
        data["id"] = group_id

        # 404 means that the LDAP group link is already removed, so let's accept it for idempotency
        self._make_requests_to_api(
            "groups/%s/ldap_group_links",
            group_id,
            method="DELETE",
            data=data,
            expected_codes=[204, 404],
        )



import pytest

from gitlabform import EXIT_INVALID_INPUT
from gitlabform.configuration import Configuration
from gitlabform.lists.filter import OmitEmptyConfigs


def test_error_on_missing_key():
    config_yaml = """
    ---
    # no key at all
    """

    with pytest.raises(SystemExit) as e:
        configuration = Configuration(config_string=config_yaml)
        OmitEmptyConfigs(configuration, None, None)
    assert e.value.code == EXIT_INVALID_INPUT


def test_error_on_empty_key():
    config_yaml = """
    ---
    projects_and_groups:
    """

    with pytest.raises(SystemExit) as e:
        configuration = Configuration(config_string=config_yaml)
        OmitEmptyConfigs(configuration, None, None)
    assert e.value.code == EXIT_INVALID_INPUT



from tests.acceptance import run_gitlabform


class TestGroupSettings:
    def test__edit_settings(self, gitlab, group):
        settings = gitlab.get_group_settings(group)
        assert settings["description"] != "foobar"

        edit_group_settings = f"""
        projects_and_groups:
          {group}/*:
            group_settings:
              description: foobar
        """

        run_gitlabform(edit_group_settings, group)

        settings = gitlab.get_group_settings(group)
        assert settings["visibility"] == "internal"

    def test__no_edit_needed(self, gitlab, group):
        settings = gitlab.get_group_settings(group)
        current_value = settings["path"]

        edit_group_settings = f"""
            projects_and_groups:
              {group}/*:
                group_settings:
                  path: {current_value}
            """

        run_gitlabform(edit_group_settings, group)

        settings = gitlab.get_group_settings(group)
        assert settings["path"] == current_value



from gitlabform.gitlab.core import GitLabCore


class GitLabPipelines(GitLabCore):
    def get_pipelines(self, project_and_group_name, branch):
        pipelines = self._make_requests_to_api(
            "projects/%s/pipelines?ref=%s",
            (project_and_group_name, branch),
        )
        return pipelines

    def get_pipeline(self, project_and_group_name, pipeline_id):
        pipeline = self._make_requests_to_api(
            "/projects/%s/pipelines/%s", (project_and_group_name, pipeline_id)
        )
        return pipeline

    def retry_pipeline(self, project_and_group_name, pipeline_id):
        pipeline = self._make_requests_to_api(
            "projects/%s/pipelines/%s/retry",
            (project_and_group_name, pipeline_id),
            method="POST",
            expected_codes=[200, 201],
        )
        return pipeline



from gitlabform.gitlab import AccessLevel


def test_group_access_levels_type():
    group_access_levels = AccessLevel.group_levels()
    assert type(group_access_levels[0]) is int


def test_group_access_levels():
    group_access_levels = AccessLevel.group_levels()
    assert AccessLevel.NO_ACCESS.value in group_access_levels
    assert AccessLevel.OWNER.value in group_access_levels
    assert AccessLevel.ADMIN.value not in group_access_levels


def test_sortability_of_access_levels():
    sorted_group_access_levels = sorted(AccessLevel)
    assert sorted_group_access_levels[0] == AccessLevel.NO_ACCESS.value
    assert sorted_group_access_levels[-1] == AccessLevel.ADMIN.value



from gitlabform.gitlab import AccessLevel
from tests.acceptance import (
    run_gitlabform,
)


class TestMembersEnforce:
    def test__enforce(
        self,
        gitlab,
        group_and_project,
        three_members,
        outsider_user,
    ):
        members_before = gitlab.get_project_members(group_and_project)
        assert len(members_before) > 0

        members_usernames_before = [member["username"] for member in members_before]
        assert outsider_user not in members_usernames_before

        enforce_users = f"""
            projects_and_groups:
              {group_and_project}:
                members:
                  users:
                    {outsider_user}: # new user
                      access_level: {AccessLevel.MAINTAINER.value}
                  enforce: true
            """

        run_gitlabform(enforce_users, group_and_project)

        members = gitlab.get_project_members(group_and_project)

        assert len(members) == 1
        assert members[0]["username"] == outsider_user



from gitlabform.gitlab.projects import GitLabProjects


class GitLabResourceGroups(GitLabProjects):
    def get_specific_resource_group(self, project_and_group_name, resource_group_name):
        return self._make_requests_to_api(
            "projects/%s/resource_groups/%s",
            (project_and_group_name, resource_group_name),
            method="GET",
            expected_codes=[200],
        )

    def update_resource_group(
        self, project_and_group_name, resource_group_name, process_mode
    ):
        return self._make_requests_to_api(
            "projects/%s/resource_groups/%s",
            (project_and_group_name, resource_group_name),
            method="PUT",
            expected_codes=[200],
            data=process_mode,
        )



from abc import ABC, abstractmethod
from types import SimpleNamespace

from cli_ui import fatal, warning
from ruamel.yaml.comments import CommentedMap
from yamlpath import Processor
from yamlpath.exceptions import YAMLPathException
from yamlpath.wrappers import ConsolePrinter

from gitlabform.constants import EXIT_INVALID_INPUT, APPROVAL_RULE_NAME
from gitlabform.configuration import Configuration
from gitlabform.gitlab import AccessLevel
from gitlabform.gitlab import GitLab


# Configuration transformers are classes which take the input configuration as YAML and change it
# to from the more user-friendly input to an output that is more applicable to passing to GitLab
# over its API.
#
# For example, we want to operate on usernames in the configuration while GitLab sometimes operates
# on user ids. Therefore, one of the transformers changes "user_id: <number>" into "user: <username>".


class ConfigurationTransformers:
    def __init__(self, gitlab: GitLab):
        self.merge_request_approvals_transformer = MergeRequestApprovalsTransformer(
            gitlab
        )
        self.user_transformer = UserTransformer(gitlab)
        self.group_transformer = GroupTransformer(gitlab)
        self.implicit_name_transformer = ImplicitNameTransformer(gitlab)
        self.access_level_transformer = AccessLevelsTransformer(gitlab)

    def transform(self, configuration: Configuration) -> None:
        # this needs to run before user and group transformer as it may contain users and group
        # names to convert to ids
        self.merge_request_approvals_transformer.transform(configuration)

        self.user_transformer.transform(configuration)
        self.group_transformer.transform(configuration)
        self.implicit_name_transformer.transform(configuration)
        self.access_level_transformer.transform(configuration)


class ConfigurationTransformer(ABC):
    @abstractmethod
    def transform(self, configuration: Configuration) -> None:
        pass


class UserTransformer(ConfigurationTransformer):
    def __init__(self, gitlab: GitLab):
        self.gitlab = gitlab

    def transform(self, configuration: Configuration) -> None:
        logging_args = SimpleNamespace(quiet=False, verbose=False, debug=False)

        processor = Processor(ConsolePrinter(logging_args), configuration.config)

        paths_to_user = [
            "projects_and_groups.*.protected_environments.*.deploy_access_levels.user"
        ]

        for path in paths_to_user:
            try:
                for node_coordinate in processor.get_nodes(path):
                    user = node_coordinate.parent.pop("user")

                    node_coordinate.parent["user_id"] = self.gitlab._get_user_id(user)
            except YAMLPathException as e:
                # this just means that we haven't found any keys in YAML
                # under the given path
                pass

        paths_to_users = ["projects_and_groups.*.merge_requests_approval_rules.*.users"]

        for path in paths_to_users:
            try:
                for node_coordinate in processor.get_nodes(path):
                    user_ids = []
                    users = node_coordinate.parent.pop("users")
                    for user in users:
                        user_id = self.gitlab._get_user_id(user)
                        user_ids.append(user_id)
                    node_coordinate.parent["user_ids"] = user_ids
            except YAMLPathException as e:
                # this just means that we haven't found any keys in YAML
                # under the given path
                pass


class GroupTransformer(ConfigurationTransformer):
    def __init__(self, gitlab: GitLab):
        self.gitlab = gitlab

    def transform(self, configuration: Configuration) -> None:
        logging_args = SimpleNamespace(quiet=False, verbose=False, debug=False)

        processor = Processor(ConsolePrinter(logging_args), configuration.config)

        paths_to_groups = [
            "projects_and_groups.*.merge_requests_approval_rules.*.groups"
        ]

        for path in paths_to_groups:
            try:
                for node_coordinate in processor.get_nodes(path):
                    group_ids = []
                    groups = node_coordinate.parent.pop("groups")
                    for group in groups:
                        group_id = self.gitlab._get_group_id(group)
                        group_ids.append(group_id)
                    node_coordinate.parent["group_ids"] = group_ids
            except YAMLPathException:
                # this just means that we haven't found any keys in YAML
                # under the given path
                pass


class ImplicitNameTransformer(ConfigurationTransformer):
    """
    Creates a 'name' field that has the same value as the "scope" delimiter, e.g.:

    ...
      blah: # start of the cfg scope
       name: blah # name to be used
       smth_else: <...>

    It's redundant, so this can be done as :

    ...
     foo: # a 'name' field will be created as -> name: foo
       smth_else: <...>
    """

    def __init__(self, gitlab: GitLab):
        # this transformer doesn't need to call gitlab
        pass

    def transform(self, configuration: Configuration) -> None:
        logging_args = SimpleNamespace(quiet=False, verbose=False, debug=False)

        processor = Processor(ConsolePrinter(logging_args), configuration.config)

        paths_to_implicit_names = ["projects_and_groups.*.protected_environments.*"]

        for path in paths_to_implicit_names:
            try:
                for node_coordinate in processor.get_nodes(path):
                    if not isinstance(node_coordinate.node, CommentedMap):
                        continue

                    node_coordinate.parent[node_coordinate.parentref][
                        "name"
                    ] = node_coordinate.parentref
            except YAMLPathException:
                # this just means that we haven't found any keys in YAML
                # under the given path
                pass


class AccessLevelsTransformer(ConfigurationTransformer):
    """
    Internally the app supports only numeric access levels, but for user-friendliness
    this class allows providing them as strings and transforms these strings into
    the appropriate numbers.
    """

    def __init__(self, gitlab: GitLab):
        # this transformer doesn't need to call gitlab
        pass

    def transform(self, configuration: Configuration):
        logging_args = SimpleNamespace(quiet=False, verbose=False, debug=False)
        log = ConsolePrinter(logging_args)

        processor = Processor(log, configuration.config)

        # [.!<100] effectively means that the value is non-numerical
        paths_to_hashes = [
            # # branches, old syntax
            "**.push_access_level[.!<100]",
            "**.merge_access_level[.!<100]",
            "**.unprotect_access_level[.!<100]",
            # members & group members
            "**.access_level[.!<100]",
            "**.group_access[.!<100]",
            # old syntax
            "**.group_access_level[.!<100]",
            # tags
            "**.create_access_level[.!<100]",
        ]

        for path in paths_to_hashes:
            try:
                for node_coordinate in processor.get_nodes(path):
                    try:
                        access_level_string = str(node_coordinate.node)
                        node_coordinate.parent[
                            node_coordinate.parentref
                        ] = AccessLevel.get_value(access_level_string)
                    except KeyError:
                        fatal(
                            f"Configuration string '{access_level_string}' is not one of the valid access levels:"
                            f" {', '.join(AccessLevel.get_canonical_names())}",
                            exit_code=EXIT_INVALID_INPUT,
                        )
            except YAMLPathException:
                # this just means that we haven't found any keys in YAML
                # under the given path
                pass

        # these are different from the above, as they are elements of arrays,
        # so we need different search query and an extra condition for
        # transformation
        paths_to_arrays = [
            # # branches, new GitLab Premium syntax
            "**.allowed_to_push.*.[access_level!<100]",
            "**.allowed_to_merge.*.[access_level!<100]",
            "**.allowed_to_unprotect.*.[access_level!<100]",
        ]

        for path in paths_to_arrays:
            try:
                for node_coordinate in processor.get_nodes(path):
                    if node_coordinate.parentref == "access_level":
                        try:
                            access_level_string = str(node_coordinate.node)
                            node_coordinate.parent[
                                node_coordinate.parentref
                            ] = AccessLevel.get_value(access_level_string)
                        except KeyError:
                            fatal(
                                f"Configuration string '{access_level_string}' is not one of the valid access levels:"
                                f" {', '.join(AccessLevel.get_canonical_names())}",
                                exit_code=EXIT_INVALID_INPUT,
                            )
            except YAMLPathException:
                # this just means that we haven't found any keys in YAML
                # under the given path
                pass


#
# TODO: remove below code in v4.x
#


class MergeRequestApprovalsTransformer(ConfigurationTransformer):
    def __init__(self, gitlab: GitLab):
        # this transformer doesn't need to call gitlab
        pass

    def transform(self, configuration: Configuration) -> None:
        logging_args = SimpleNamespace(quiet=False, verbose=False, debug=False)

        processor = Processor(ConsolePrinter(logging_args), configuration.config)

        try:
            for node_coordinate in processor.get_nodes(
                "projects_and_groups.*.merge_requests"
            ):
                old_syntax_found = False
                approvals_required_guessed = False

                where_to_add_new_syntax = node_coordinate.parent

                old_syntax = node_coordinate.node

                # approval rules
                if "approvers" in old_syntax or "approver_groups" in old_syntax:
                    old_syntax_found = True

                    if (
                        "approvals" in old_syntax
                        and "approvals_before_merge" in old_syntax["approvals"]
                    ):
                        approvals_required = old_syntax["approvals"][
                            "approvals_before_merge"
                        ]
                    else:
                        approvals_required = 2
                        approvals_required_guessed = True

                    where_to_add_new_syntax["merge_requests_approval_rules"] = {
                        "legacy": {
                            "approvals_required": approvals_required,
                            "name": APPROVAL_RULE_NAME,
                        }
                    }

                    if "approvers" in old_syntax:
                        where_to_add_new_syntax["merge_requests_approval_rules"][
                            "legacy"
                        ]["users"] = old_syntax["approvers"]
                    if "approver_groups" in old_syntax:
                        where_to_add_new_syntax["merge_requests_approval_rules"][
                            "legacy"
                        ]["groups"] = old_syntax["approver_groups"]

                    if (
                        "remove_other_approval_rules" in old_syntax
                        and old_syntax["remove_other_approval_rules"]
                    ):
                        old_syntax_found = True
                        where_to_add_new_syntax["merge_requests_approval_rules"][
                            "enforce"
                        ] = True

                # approvals configurations
                if "approvals" in old_syntax:
                    old_syntax_found = True
                    if "approvals_before_merge" in old_syntax["approvals"]:
                        # this setting is now moved inside the approval rules, so remove it from here
                        old_syntax["approvals"].pop("approvals_before_merge")
                    where_to_add_new_syntax["merge_requests_approvals"] = old_syntax[
                        "approvals"
                    ]

                if old_syntax_found:
                    where_to_add_new_syntax.pop("merge_requests")

                if old_syntax_found:
                    warning(
                        "The 'merge_requests' configuration section syntax works but is deprecated and will be removed "
                        "in the next major version of GitLabForm. "
                        "Please migrate to the new syntax using the 'merge_requests_approvals' "
                        "and the 'merge_requests_approval_rules' sections."
                    )
                if approvals_required_guessed:
                    warning(
                        "'approvals_before_merge' not found in the old 'merge_requests' configuration section",
                        "- assuming that it is set to 2.",
                    )

        except YAMLPathException:
            # this just means that we haven't found any keys in YAML
            # under the given path
            pass
